# noinspection PyUnresolvedReferences,PyPackageRequirements
from pytrain import PowerDistrict, GantryCrane, Switch

g = GantryCrane(
    address=87,
    cab_left_pin=20,
    cab_right_pin=21,
    ro_left_pin=12,
    ro_right_pin=16,
    bo_down_pin=8,
    bo_up_pin=7,
    mag_pin=23,
    led_pin=24,
    cab_rotary_encoder=True,
)

pd13 = PowerDistrict(
    13,
    on_pin=17,
    off_pin=27,
    on_led_pin=22,
)

pd14 = PowerDistrict(
    14,
    on_pin=10,
    off_pin=9,
    on_led_pin=11,
)

sw1 = Switch(
    1,
    thru_pin=5,
    out_pin=6,
    thru_led_pin=13,
    out_led_pin=19,
)



from __future__ import annotations

import logging
from abc import ABC, ABCMeta

from .sequence_constants import SequenceCommandEnum
from .sequence_req import SequenceReq, T
from ..constants import CommandScope
from ..tmcc1.tmcc1_constants import TMCC1EngineCommandEnum
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum, tmcc2_speed_to_rpm
from ...db.component_state_store import ComponentStateStore

log = logging.getLogger(__name__)


def labor_delta(cur_speed: int, new_speed: int, cur_labor: int) -> int:
    delta = new_speed - cur_speed
    if delta > 0:
        return min(31, max(0, round((delta * 0.09546) - 0.5401)) + cur_labor)
    elif delta < 0:
        return max(0, cur_labor - max(0, round((-0.06030 * delta) + 0.01052)))
    else:
        return cur_labor


class RampedSpeedReqBase(SequenceReq, ABC):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        command: SequenceCommandEnum,
        address: int,
        speed: int | str | T = None,
        scope: CommandScope = CommandScope.ENGINE,
        dialog: bool = False,
        is_tmcc: bool = False,
    ) -> None:
        super().__init__(command, address, scope)
        tower, s, speed_req, engr = self.decode_rr_speed(speed, is_tmcc)
        # if an integer speed was provided, use it, otherwise, rely on rr speed
        # # provided by decode call; only do this if dialogs are NOT requested
        if isinstance(speed, int) and dialog is False:
            speed_req = speed
        # get current state record
        cur_state = ComponentStateStore.get_state(scope, address, create=False)
        # if there is no state information, treat this as an ABSOLUTE_SPEED req
        if cur_state is None or cur_state.speed is None:
            if tower and engr and dialog is True:
                from .speed_req import SpeedReq

                sr = SpeedReq(address, speed, scope, is_tmcc)
                for request in sr.requests:
                    self.add(request.request, delay=request.delay, repeat=request.repeat)
            else:
                speed_enum = TMCC1EngineCommandEnum.ABSOLUTE_SPEED if is_tmcc else TMCC2EngineCommandEnum.ABSOLUTE_SPEED
                self.add(speed_enum, address, speed_req, scope)
                if is_tmcc is False:
                    rpm = tmcc2_speed_to_rpm(speed_req)
                    self.add(TMCC2EngineCommandEnum.DIESEL_RPM, address, data=rpm, scope=scope, delay=4)
        else:
            speed_enum = (
                TMCC2EngineCommandEnum.ABSOLUTE_SPEED if cur_state.is_legacy else TMCC1EngineCommandEnum.ABSOLUTE_SPEED
            )
            # issue tower dialog, if requested
            if tower and dialog is True:
                self.add(tower, address, scope=scope)
            # use current speed and momentum to build up or down speed
            cs = cur_state.speed
            delay = 0.0
            inc = 3
            c_rpm = cur_state.rpm
            init_labor = cur_state.labor
            if cur_state.momentum is not None:
                delay_inc = 0.200 + (cur_state.momentum * 0.010)
                inc = 2 if cur_state.momentum >= 6 else inc
                inc = 1 if cur_state.momentum >= 7 else inc
            else:
                delay_inc = 0.200
            speed_req = min(speed_req, cur_state.speed_max)
            # are we speeding up or down?
            log.debug(f"CS: {cs} Requested Speed: {speed_req}")
            ramp = range(cs + inc, speed_req + 1, inc) if cs < speed_req else range(cs - inc, speed_req + 1, -inc)
            if ramp:
                # increase or decrease labor
                c_labor = labor_delta(cs, speed_req, init_labor)
                self.add(TMCC2EngineCommandEnum.ENGINE_LABOR, address, data=c_labor, scope=scope, delay=delay)
                # if we're decelerating, kill RPM and labor up front
                if cur_state.is_legacy and cs > speed_req:
                    if cur_state.is_rpm:
                        rpm = tmcc2_speed_to_rpm(speed_req)
                        self.add(TMCC2EngineCommandEnum.DIESEL_RPM, address, data=rpm, scope=scope, delay=delay)
                        c_rpm = rpm
                for speed in ramp:
                    self.add(speed_enum, address, speed, scope, delay=delay)
                    if cur_state.is_legacy:
                        labor = labor_delta(speed, speed_req, init_labor)
                        if labor != c_labor:
                            self.add(TMCC2EngineCommandEnum.ENGINE_LABOR, address, data=labor, scope=scope, delay=delay)
                            c_labor = labor
                        if cur_state.is_rpm and cs < speed_req:
                            rpm = tmcc2_speed_to_rpm(speed)
                            if rpm != c_rpm:
                                self.add(TMCC2EngineCommandEnum.DIESEL_RPM, address, data=rpm, scope=scope, delay=delay)
                                c_rpm = rpm
                    delay += delay_inc
                # make sure the final speed is requested
                if ramp[-1] != speed_req:
                    self.add(speed_enum, address, speed_req, scope, delay=delay)
                    if cur_state.is_legacy:
                        self.add(
                            TMCC2EngineCommandEnum.ENGINE_LABOR, address, data=init_labor, scope=scope, delay=delay
                        )
                        if cur_state.is_rpm:
                            rpm = tmcc2_speed_to_rpm(speed_req)
                            if rpm != c_rpm:
                                self.add(TMCC2EngineCommandEnum.DIESEL_RPM, address, data=rpm, scope=scope, delay=delay)
            else:
                self.add(speed_enum, address, speed_req, scope)
            # issue engineer dialog, if requested
            if engr and dialog is True:
                if delay < 2.00:
                    delay = 2.50
                self.add(engr, address, scope=scope, delay=delay)


class RampedSpeedReq(RampedSpeedReqBase):
    def __init__(
        self,
        address: int,
        speed: int | str | T = None,
        scope: CommandScope = CommandScope.ENGINE,
        is_tmcc: bool = False,
    ) -> None:
        super().__init__(SequenceCommandEnum.RAMPED_SPEED_SEQ, address, speed, scope, is_tmcc=is_tmcc)


SequenceCommandEnum.RAMPED_SPEED_SEQ.value.register_cmd_class(RampedSpeedReq)


class RampedSpeedDialogReq(RampedSpeedReqBase):
    def __init__(
        self,
        address: int,
        speed: int | str | T = None,
        scope: CommandScope = CommandScope.ENGINE,
        is_tmcc: bool = False,
    ) -> None:
        super().__init__(
            SequenceCommandEnum.RAMPED_SPEED_DIALOG_SEQ,
            address,
            speed,
            scope,
            dialog=True,
            is_tmcc=is_tmcc,
        )


SequenceCommandEnum.RAMPED_SPEED_DIALOG_SEQ.value.register_cmd_class(RampedSpeedDialogReq)



from __future__ import annotations

from typing import Dict, TypeVar

from ..command_def import CommandDefEnum
from ..constants import CommandScope, CommandSyntax
from ..tmcc2.tmcc2_constants import TMCC2CommandDef

S = TypeVar("S", bound="SequenceDef")


class SequenceDef(TMCC2CommandDef):
    def __init__(
        self,
        command_bits: int,
        scope: CommandScope = CommandScope.ENGINE,
        is_addressable: bool = True,
        d_min: int = 0,
        d_max: int = 0,
        d_map: Dict[int, int] = None,
        interval: int = None,
        cmd_class: S = None,
    ) -> None:
        super().__init__(
            command_bits,
            scope=scope,
            is_addressable=is_addressable,
            d_min=d_min,
            d_max=d_max,
            d_map=d_map,
            interval=interval,
        )
        self._cmd_class = cmd_class

    @property
    def cmd_class(self) -> S:
        return self._cmd_class

    def register_cmd_class(self, cmd_class: S) -> None:
        if self._cmd_class is None:
            self._cmd_class = cmd_class

    @property
    def syntax(self) -> CommandSyntax:
        return CommandSyntax.LEGACY

    @property
    def is_tmcc1(self) -> bool:
        return self.syntax == CommandSyntax.TMCC

    @property
    def is_tmcc2(self) -> bool:
        return self.syntax == CommandSyntax.LEGACY

    @property
    def first_byte(self) -> bytes | None:
        raise NotImplementedError

    @property
    def address_mask(self) -> bytes | None:
        raise NotImplementedError


class SequenceCommandEnum(CommandDefEnum):
    ABSOLUTE_SPEED_SEQ = SequenceDef(1, d_max=199)
    RAMPED_SPEED_SEQ = SequenceDef(2, d_max=199)
    RAMPED_SPEED_DIALOG_SEQ = SequenceDef(3, d_max=199)
    GRADE_CROSSING_SEQ = SequenceDef(4)
    LABOR_EFFECT_DOWN_SEQ = SequenceDef(5)
    LABOR_EFFECT_UP_SEQ = SequenceDef(6)
    ABSOLUTE_SPEED_RPM = SequenceDef(7, d_max=199)



from __future__ import annotations

from ..pdi.constants import PdiCommand, Bpc2Action, PDI_SOP, PDI_EOP
from ..pdi.lcs_req import LcsReq
from ..protocol.constants import CommandScope


class Bpc2Req(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.BPC2_GET,
        action: Bpc2Action = Bpc2Action.CONFIG,
        ident: int | None = None,
        error: bool = False,
        mode: int = None,
        debug: int = None,
        restore: bool = None,
        state: int = None,
        values: int = None,
        valids: int = None,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        if isinstance(data, bytes):
            self._action = Bpc2Action(self._action_byte)
            data_len = len(self._data)
            if self._action == Bpc2Action.CONFIG:
                self._debug = self._data[4] if data_len > 4 else None
                self._mode = self._data[7] if data_len > 7 else None
                self._restore = (self._mode & 0x80) == 0x80
                if self._restore:
                    self._mode &= 0x7F
            else:
                self._mode = self._debug = self._restore = None

            if self._action in {Bpc2Action.CONTROL1, Bpc2Action.CONTROL3}:
                self._state = self._data[3] if data_len > 3 else None
                self._values = self._valids = None
                if self._action == Bpc2Action.CONTROL3:
                    self.scope = CommandScope.ACC
                else:
                    self.scope = CommandScope.TRAIN
            elif self._action in {Bpc2Action.CONTROL2, Bpc2Action.CONTROL4}:
                self._values = self._data[3] if data_len > 3 else None
                self._valids = self._data[4] if data_len > 4 else None
                self._state = None
                if self._action == Bpc2Action.CONTROL4:
                    self.scope = CommandScope.ACC
                else:
                    self.scope = CommandScope.TRAIN
            else:
                self._state = self._values = self._valids = None
        else:
            self._action = action
            self._mode = mode
            self._debug = debug
            self._restore = restore
            self._state = state
            self._values = values
            self._valids = valids

    @property
    def mode(self) -> int | None:
        return self._mode

    @property
    def restore(self) -> bool:
        return self._restore

    @property
    def debug(self) -> int | None:
        return self._debug

    @property
    def state(self) -> int | None:
        return self._state

    @property
    def values(self) -> int | None:
        return self._values

    @property
    def valids(self) -> int | None:
        return self._valids

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        if self.pdi_command != PdiCommand.BPC2_GET:
            if self.action == Bpc2Action.CONFIG:
                return f"Mode: {self.mode} Debug: Restore: {self.restore} {self.debug} ({self.packet})"
            elif self.action in {Bpc2Action.CONTROL1, Bpc2Action.CONTROL3}:
                return f"Power {'ON' if self.state == 1 else 'OFF'} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += self.action.as_bytes

        if self._action == Bpc2Action.CONFIG:
            if self.pdi_command != PdiCommand.BPC2_GET:
                debug = self.debug if self.debug is not None else 0
                mode = self.mode | 0x80 if self.restore else self.mode
                byte_str += self.tmcc_id.to_bytes(1, byteorder="big")  # allows board to be renumbered
                byte_str += debug.to_bytes(1, byteorder="big")
                byte_str += (0x0000).to_bytes(2, byteorder="big")
                byte_str += mode.to_bytes(1, byteorder="big")
        elif self._action == Bpc2Action.IDENTIFY:
            if self.pdi_command == PdiCommand.BPC2_SET:
                byte_str += (self.ident if self.ident is not None else 0).to_bytes(1, byteorder="big")
        elif self._action in {Bpc2Action.CONTROL1, Bpc2Action.CONTROL3}:
            if self.pdi_command != PdiCommand.BPC2_GET:
                byte_str += (self.state if self.state is not None else 0).to_bytes(1, byteorder="big")
        elif self._action in {Bpc2Action.CONTROL2, Bpc2Action.CONTROL4}:
            if self.pdi_command != PdiCommand.BPC2_GET:
                byte_str += (self.state if self.state is not None else 0).to_bytes(1, byteorder="big")
                values = self.values if self.values is not None else 0
                valids = self.values if self.valids is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += valids.to_bytes(1, byteorder="big")
        elif self._action == Bpc2Action.CONTROL2:
            if self.pdi_command != PdiCommand.BPC2_GET:
                values = self.values if self.values is not None else 0
                valids = self.values if self.valids is not None else 0
                byte_str += values.to_bytes(1, byteorder="big")
                byte_str += valids.to_bytes(1, byteorder="big")
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str



#
#
# PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
# Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
# SPDX-License-Identifier: LPGL

import importlib.metadata
import platform
import sys
from importlib.metadata import PackageNotFoundError

from .atc.block import Block  # noqa: F401
from .cli.pytrain import (
    PyTrain,
    PyTrainExitException,  # noqa: F401
    PyTrainExitStatus,  # noqa: F401
)
from .db.component_state import (
    ComponentState,  # noqa: F401
    RouteState,  # noqa: F401
    SwitchState,  # noqa: F401
)
from .db.accessory_state import AccessoryState  # noqa: F401
from .db.engine_state import EngineState, TrainState  # noqa: F401
from .db.irda_state import IrdaState  # noqa: F401
from .db.sync_state import SyncState  # noqa: F401
from .db.component_state_store import ComponentStateStore  # noqa: F401
from .gpio.base_watcher import BaseWatcher  # noqa: F401
from .gpio.controller import Controller  # noqa: F401
from .gpio.crane_car import CraneCar  # noqa: F401
from .gpio.culvert_loader import CulvertLoader, CulvertUnloader  # noqa: F401
from .gpio.engine_controller import EngineController  # noqa: F401
from .gpio.engine_status import EngineStatus  # noqa: F401
from .gpio.gantry_crane import GantryCrane  # noqa: F401
from .gpio.gpio_handler import (
    GpioHandler,  # noqa: F401
    JoyStickHandler,  # noqa: F401
    PotHandler,  # noqa: F401
)
from .gpio.launch_pad import LaunchPad  # noqa: F401
from .gpio.launch_status import LaunchStatus  # noqa: F401
from .gpio.power_district import PowerDistrict  # noqa: F401
from .gpio.power_watcher import PowerWatcher  # noqa: F401
from .gpio.route import Route  # noqa: F401
from .gpio.smoke_fluid_loader import SmokeFluidLoader  # noqa: F401
from .gpio.switch import Switch  # noqa: F401
from .gpio.sys_admin import SystemAdmin  # noqa: F401
from .protocol.command_def import CommandDefEnum  # noqa: F401
from .protocol.command_req import CommandReq  # noqa: F401
from .protocol.constants import (
    PROGRAM_NAME,
    CommandScope,  # noqa: F401
    CommandSyntax,  # noqa: F401
    ControlType,  # noqa: F401
)
from .protocol.multibyte.multibyte_constants import (
    TMCC2EffectsControl,  # noqa: F401
    TMCC2LightingControl,  # noqa: F401
    TMCC2MaskingControl,  # noqa: F401
    TMCC2R4LCEnum,  # noqa: F401
    TMCC2RailSoundsDialogControl,  # noqa: F401
    TMCC2RailSoundsEffectsControl,  # noqa: F401
    TMCC2VariableEnum,  # noqa: F401
    UnitAssignment,  # noqa: F401
)
from .protocol.sequence.sequence_constants import (
    SequenceCommandEnum,  # noqa: F401
)
from .protocol.sequence.sequence_req import (
    SequencedReq,  # noqa: F401
    SequenceReq,  # noqa: F401
)
from .protocol.tmcc1.tmcc1_constants import (
    TMCC1AuxCommandEnum,  # noqa: F401
    TMCC1EngineCommandEnum,  # noqa: F401
    TMCC1HaltCommandEnum,  # noqa: F401
    TMCC1RouteCommandEnum,  # noqa: F401
    TMCC1RRSpeedsEnum,  # noqa: F401
    TMCC1SwitchCommandEnum,  # noqa: F401
)
from .protocol.tmcc2.tmcc2_constants import (
    TMCC2EngineCommandEnum,  # noqa: F401
    TMCC2HaltCommandEnum,  # noqa: F401
    TMCC2RouteCommandEnum,  # noqa: F401
    TMCC2RRSpeedsEnum,  # noqa: F401
)
from .utils.path_utils import (
    find_dir,  # noqa: F401
    find_file,  # noqa: F401
)

PROGRAM_PACKAGE = "pytrain-ogr"


def main(args: list[str] | None = None) -> int:
    if args is None:
        args = sys.argv[1:]
    try:
        PyTrain(args)
        return 0
    except Exception as e:
        # Output anything else nicely formatted on stderr and exit code 1
        sys.exit(f"{PROGRAM_NAME}: error: {e}\n")


def is_package() -> bool:
    try:
        # production version
        importlib.metadata.version(PROGRAM_PACKAGE)
        return True
    except PackageNotFoundError:
        return False


def is_linux() -> bool:
    return platform.system().lower() == "linux"


def get_version() -> str:
    #
    # this should be easier, but it is what it is.
    # we handle the two major cases; we're running from
    # the PyTrain pypi package, or we're running from
    # source retrieved from git...
    #
    # we try the package path first...
    version = None
    try:
        # production version
        version = importlib.metadata.version(PROGRAM_PACKAGE)
    except PackageNotFoundError:
        pass

    # finally, call the method to read it from git
    if version is None:
        from setuptools_scm import get_version as get_git_version

        version = get_git_version(version_scheme="only-version")

    version = version if version.startswith("v") else f"v{version}"
    version = version.replace(".post0", "")
    return version


def get_version_tuple() -> tuple[int, int, int]:
    version = get_version()
    version = version.replace("v", "")
    if "+" in version:
        plus_pos = version.find("+")
        version = version[0:plus_pos]
    version = version.split(".")
    return int(version[0]), int(version[1]), int(version[2])


def get_version_bytes() -> bytes:
    version = get_version_tuple()
    ver_bytes = bytes()
    for v in version:
        ver_bytes += v.to_bytes(1, "big")
    return ver_bytes


SMOKE_LEVEL_MAP = {
    0: TMCC2EffectsControl.SMOKE_OFF,
    1: TMCC2EffectsControl.SMOKE_LOW,
    2: TMCC2EffectsControl.SMOKE_MEDIUM,
    3: TMCC2EffectsControl.SMOKE_HIGH,
}



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from threading import RLock
from time import time


class ExpiringSet:
    def __init__(self, max_age_seconds=1.0):
        assert max_age_seconds > 0
        self.age = max_age_seconds
        self.container = dict()
        self._lock = RLock()

    def __repr__(self) -> str:
        rep = ""
        for k, v in self.container.items():
            ttl = max(self.age - time() - v, 0)
            if ttl > 0:
                rep += f"{k.hex()} TTL: {ttl}\n"
        return rep

    def __len__(self) -> int:
        return len(self.container)

    def __contains__(self, value) -> bool:
        return self.contains(value)

    def contains(self, value):
        with self._lock:
            if value not in self.container:
                return False
            if time() - self.container[value] > self.age:
                del self.container[value]
                return False
            return True

    def add(self, value):
        with self._lock:
            if self.contains(value) is False:
                self.container[value] = time()

    def clear(self):
        with self._lock:
            self.container.clear()

    def discard(self, value):
        with self._lock:
            if value in self.container:
                del self.container[value]

    def remove(self, value):
        with self._lock:
            if value not in self.container:
                raise KeyError(f"{value} not found in set")
            self.discard(value)



import abc
from abc import ABC

from .tmcc1_constants import TMCC1_COMMAND_PREFIX

from ..command_base import CommandBase
from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT
from ..constants import CommandScope
from ..command_def import CommandDefEnum


class TMCC1Command(CommandBase, ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        command: CommandDefEnum,
        command_req: CommandReq,
        address: int = 99,
        data: int = 0,
        scope: CommandScope = None,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        super().__init__(command, command_req, address, data, scope, baudrate, port, server)

    def __repr__(self):
        name = self._command_def_enum.name
        data = f" [{self._command_req.data}] " if self._command_req.num_data_bits else ""
        entity = {self._command_req.scope.name.title(): self._command_req.scope.name}
        return f"<{name}: {entity} {self.address} {data}: 0x{self.command_bytes.hex()}>"

    def _encode_address(self, command_op: int) -> bytes:
        return self._encode_command((self.address << 7) | command_op)

    def _command_prefix(self) -> bytes:
        return TMCC1_COMMAND_PREFIX.to_bytes(1, "big")

    def _build_command(self) -> bytes:
        return self._command_req.as_bytes



from __future__ import annotations

from typing import Dict

from .constants import PdiCommand, IrdaAction, PDI_SOP, PDI_EOP
from .lcs_req import LcsReq
from ..protocol.constants import CommandScope, Mixins


class IrdaSequence(Mixins):
    NONE = 0
    CROSSING_GATE_NONE = 1
    NONE_CROSSING_GATE = 2
    BELL_NONE = 3
    NONE_BELL = 4
    JOURNEY_BEGIN_JOURNEY_END = 5
    JOURNEY_END_JOURNEY_BEGIN = 6
    SLOW_SPEED_NORMAL_SPEED = 7
    NORMAL_SPEED_SLOW_SPEED = 8
    RECORDING = 9


STATUS_MAP: Dict[int, str] = {
    0: "No Recording",
    1: "Idle",
    2: "Playback...",
    3: "Armed...",
    4: "Recording...",
}

PRODUCT_REV_MAP: Dict[int, str] = {
    0x00: "Switcher",
    0x01: "Road",
}

PRODUCT_ID_MAP: Dict[int, str] = {
    0x02: "Diesel",
    0x03: "Diesel Switcher",
    0x04: "Steam",
    0x05: "Steam Switcher",
    0x06: "Subway",
    0x07: "Electric",
    0x08: "Acela",
    0x09: "Pullmor Diesel",
    0x0A: "Pullmor Steam",
    0x0B: "Breakdown",
    0x0C: "Track Crane",
    0x0D: "Accessory",
    0x0E: "Stock Car",
    0x0F: "Passenger Car",
}

TSDB_MAP: Dict[int, str] = {
    0x01: "Ditch Lights",
    0x02: "Ground Lights",
    0x03: "MARS Lights",
    0x04: "Hazard Lights",
    0x05: "Strobe Lights",
    0x06: "Reserved",
    0x07: "Reserved",
    0x08: "Rule 17",
    0x09: "Loco Marker",
    0x0A: "Tender Marker",
    0x0B: "Doghouse",
    0x0C: "Reserved",
    0x0D: "Reserved",
    0x0E: "Reserved",
    0x0F: "Reserved",
}


class IrdaReq(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.IRDA_GET,
        action: IrdaAction = IrdaAction.CONFIG,
        scope: CommandScope = CommandScope.IRDA,
        ident: int | None = None,
        error: bool = False,
        sequence: IrdaSequence | int = None,
        debug: int = 0,
        loco_rl: int = 255,
        loco_lr: int = 255,
    ) -> None:
        self._debug = self._sequence = self._loco_rl = self._loco_lr = None
        self._valid1 = self._valid2 = self._dir = self._engine_id = self._train_id = self._status = None
        self._fuel = self._water = self._burn = self._fwb_mask = None
        super().__init__(data, pdi_command, action, ident, error)
        self.scope = CommandScope.IRDA
        if isinstance(data, bytes):
            self._action = IrdaAction(self._action_byte)
            data_len = len(self._data)
            if self._action == IrdaAction.INFO:
                self.scope = CommandScope.ACC
            elif self._action == IrdaAction.CONFIG:
                self._debug = self._data[4] if data_len > 4 else None
                self._sequence = IrdaSequence.by_value(self._data[7], True) if data_len > 7 else None
                self._loco_rl = self._data[8] if data_len > 8 else None
                self._loco_lr = self._data[9] if data_len > 9 else None
            elif self._action == IrdaAction.DATA:
                self._valid1 = int.from_bytes(self._data[3:4], byteorder="little") if data_len > 4 else None
                self._valid2 = int.from_bytes(self._data[5:6], byteorder="little") if data_len > 6 else None
                self._dir = self._data[7] if data_len > 7 else None
                self._engine_id = self._data[8] if data_len > 8 else None
                self._train_id = self._data[9] if data_len > 9 else None
                self._status = self._data[10] if data_len > 10 else None
                self._fuel = self._data[11] if data_len > 11 else None
                self._water = self._data[12] if data_len > 12 else None
                self._burn = self._data[13] if data_len > 13 else None
                self._fwb_mask = self._data[14] if data_len > 14 else None
                self._runtime = int.from_bytes(self._data[15:16], byteorder="little") if data_len > 15 else None
                self._prod_rev = self._data[17] if data_len > 17 else None
                self._prod_id = self._data[18] if data_len > 18 else None
                self._bluetooth_id = self._data[19:21] if data_len > 20 else None
                self._prod_year = 2000 + int(self.decode_text(self._data[21:])) if data_len > 21 else None
                self._name = self.decode_text(self._data[24:58]) if data_len > 24 else None
                self._number = self.decode_text(self._data[57:]) if data_len > 58 else None
                self._tsdb_left = self._data[62] if data_len > 62 else None
                self._tsdb_right = self._data[63] if data_len > 63 else None
                self._max_speed = self._data[64] if data_len > 64 else None
                self._odometer = int.from_bytes(self._data[65:68], byteorder="little") if data_len > 68 else None
            elif self._action == IrdaAction.SEQUENCE:
                self._sequence = IrdaSequence.by_value(self._data[3], True) if data_len > 3 else None
            elif self._action == IrdaAction.RECORD:
                self._status = self._data[3] if data_len > 3 else None
        else:
            self.scope = scope if scope is not None else CommandScope.IRDA
            self._debug = debug
            self._sequence = (
                sequence
                if sequence is None or isinstance(sequence, IrdaSequence)
                else IrdaSequence.by_value(sequence, True)
            )
            self._loco_rl = loco_rl
            self._loco_lr = loco_lr

    @property
    def debug(self) -> int | None:
        return self._debug

    @property
    def is_right_to_left(self) -> bool:
        return True if self._dir == 0 else False

    @property
    def is_left_to_right(self) -> bool:
        return True if self._dir == 1 else False

    @property
    def direction(self) -> int:
        return self._dir

    @property
    def sequence(self) -> IrdaSequence:
        return self._sequence

    @property
    def sequence_str(self) -> str | None:
        return self.sequence.name.title() if self._sequence else "NA"

    @property
    def sequence_id(self) -> int:
        return int(self.sequence.value) if self._sequence else None

    @property
    def loco_rl(self) -> int:
        return self._loco_rl

    @property
    def loco_lr(self) -> int:
        return self._loco_lr

    @property
    def train_id(self) -> int:
        return self._train_id

    @property
    def engine_id(self) -> int:
        return self._engine_id

    @property
    def bluetooth_id(self) -> bytes:
        return self._bluetooth_id

    @property
    def year(self) -> int:
        return self._prod_year

    @property
    def status(self) -> str | None:
        return STATUS_MAP[self._status] if self._status in STATUS_MAP else "NA"

    @property
    def product_rev(self) -> str | None:
        return PRODUCT_REV_MAP[self._prod_rev] if self._prod_rev in PRODUCT_REV_MAP else "NA"

    @property
    def product_id(self) -> str | None:
        return PRODUCT_ID_MAP[self._prod_id] if self._prod_id in PRODUCT_ID_MAP else "NA"

    @staticmethod
    def tsdb(code: int) -> str | None:
        return TSDB_MAP[code] if code in TSDB_MAP else "<Blank>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def number(self) -> str:
        return self._number

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        if self.pdi_command != PdiCommand.IRDA_GET:
            if self.action == IrdaAction.CONFIG:
                rle = f"{self._loco_rl}" if self._loco_rl and self._loco_rl != 255 else "Any"
                lre = f"{self._loco_lr}" if self._loco_lr and self._loco_lr != 255 else "Any"
                rl = f" When Engine ID (R -> L): {rle}"
                lr = f" When Engine ID (L -> R): {lre}"
                return f"Sequence: {self.sequence_str}{rl}{lr} Debug: {self.debug} ({self.packet})"
            elif self.action == IrdaAction.DATA:
                trav = "R -> L" if self._dir == 0 else "L -> R"
                if self.train_id:
                    eng = f"Train: {self.train_id}"
                elif self.engine_id:
                    eng = f"Engine: {self.engine_id}"
                else:
                    eng = "<Unknown>"
                na = f" {self._name}" if self._name is not None else ""
                no = f" #{self._number}" if self._number is not None else ""
                yr = f" {self.year}" if self.year is not None else ""
                bt = f" BT: {self.bluetooth_id.hex(':')}" if self._bluetooth_id else ""
                ty = f" Type: {self.product_id}"
                ft = f" Od: {self._odometer:,} ft" if self._odometer is not None else ""
                fl = f" Fuel: {(100.0 * self._fuel / 255):.2f}%" if self._fuel is not None else ""
                wl = f" Water: {(100.0 * self._water / 255):.2f}%" if self._water is not None else ""
                return f"{trav} {eng}{na}{no}{bt}{yr}{ty}{ft}{fl}{wl} Status: {self.status} ({self.packet})"
            elif self.action == IrdaAction.SEQUENCE:
                return f"Sequence: {self.sequence_str} ({self.packet})"
            elif self.action == IrdaAction.RECORD:
                return f"Status: {self.status} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += self.action.as_bytes
        bs_len = len(byte_str)
        if self._action == IrdaAction.CONFIG:
            if self.pdi_command == PdiCommand.IRDA_RX:
                debug = self.debug if self.debug is not None else 0
                byte_str += self.tmcc_id.to_bytes(1, byteorder="big")  # allows board to be renumbered
                byte_str += debug.to_bytes(1, byteorder="big")
                byte_str += (0x0000).to_bytes(2, byteorder="big")
                byte_str += self.sequence_id.to_bytes(1, byteorder="big")
                byte_str += self.loco_rl.to_bytes(1, byteorder="big")
                byte_str += self.loco_lr.to_bytes(1, byteorder="big")
        elif self._action == IrdaAction.SEQUENCE:
            if self.pdi_command == PdiCommand.IRDA_SET:
                if self._sequence is not None:
                    byte_str += self._sequence.value.to_bytes(1, byteorder="big")
        elif self._action == IrdaAction.IDENTIFY:
            if self.pdi_command == PdiCommand.IRDA_SET:
                byte_str += (self.ident if self.ident is not None else 0).to_bytes(1, byteorder="big")
        if len(byte_str) > bs_len:
            byte_str, checksum = self._calculate_checksum(byte_str)
            byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
            byte_str += checksum
            byte_str += PDI_EOP.to_bytes(1, byteorder="big")
            return byte_str
        # if byte_str wasn't modified, return the superclass
        return super().as_bytes



from __future__ import annotations

import abc
from abc import ABC
from typing import List, TypeVar, Dict

from ..pdi.constants import (
    ALL_FIRMWARE,
    ALL_IDENTIFY,
    ALL_INFO,
    ALL_SETs,
    ALL_STATUS,
    PdiAction,
    PdiCommand,
    Ser2Action,
)
from ..pdi.pdi_req import PdiReq

T = TypeVar("T", bound=PdiAction)

BASE_TYPE_MAP: Dict[int, str] = {
    0: "Unknown",
    1: "Legacy",
    2: "Base1-L",
    3: "No Base",
    4: "No Support",
}

UART_MAP: Dict[int, str] = {
    0: "Unknown",
    1: ">Base<",
    2: "->Base",
    3: "Base->",
}

ERROR_CODE_MAP: Dict[int, str] = {
    1: "PDI command not supported",
    2: "Action not supported",
    3: "Data field with missing parameter(s) - missing bytes before EOP",
    4: "Data field with extra parameter(s) - extra bytes before EOP",
    5: "Data field with invalid contents",
    0xFF: "Undefined error",
}


class LcsReq(PdiReq, ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = None,
        action: T = None,
        ident: int | None = None,
        error: bool = False,
    ) -> None:
        super().__init__(data, pdi_command)
        self._board_id = self._num_ids = self._model = self._uart0 = self._uart1 = self._base_type = None
        self._error_code = None
        self._dc_volts: float | None = None
        self._action: T = action
        self._version = self._revision = self._sub_revision = None
        self._error = error
        if isinstance(data, bytes):
            if self.is_lcs is False:
                raise AttributeError(f"Invalid PDI LCS Request: {data}")
            self.tmcc_id = self._data[1]
            self._action_byte = self._data[2]
            if self._action_byte & 0x80 == 0x80:
                self._error = True
                self._action_byte &= 0x7F
            self._ident = None
            payload = self._data[3:]
            payload_len = len(payload)
            if self.is_error:
                self._error_code = payload[0] if payload_len > 0 else None
            if self._is_action(ALL_STATUS):
                self._board_id = payload[0] if payload_len > 0 else None
                self._num_ids = payload[1] if payload_len > 1 else None
                self._model = payload[2] if payload_len > 2 else None
                self._uart0 = payload[3] if payload_len > 3 else None
                self._uart1 = payload[4] if payload_len > 4 else None
                self._base_type = payload[5] if payload_len > 5 else None
                self._dc_volts = payload[6] / 10.0 if payload_len > 6 else None
                self._ec0 = payload[7] if payload_len > 7 else None
                self._ec1 = payload[8] if payload_len > 8 else None
                self._ec2 = payload[9] if payload_len > 9 else None
                self._ec3 = payload[10] if payload_len > 10 else None
                self._ec4 = payload[11] if payload_len > 11 else None
                self._ec5 = payload[12] if payload_len > 12 else None
                self._ec6 = payload[13] if payload_len > 13 else None
                self._ec7 = payload[14] if payload_len > 14 else None
                self._ec8 = payload[15] if payload_len > 15 else None
                self._ec9 = payload[16] if payload_len > 16 else None
                self._ec10 = payload[17] if payload_len > 17 else None
                self._ec11 = payload[18] if payload_len > 18 else None
                self._ec12 = payload[19] if payload_len > 19 else None
                self._ec13 = payload[20] if payload_len > 20 else None
                self._ec14 = payload[21] if payload_len > 21 else None
                self._ec15 = payload[22] if payload_len > 22 else None
            if self._is_action(ALL_FIRMWARE):
                self._version = payload[0] if payload_len > 0 else None
                self._revision = payload[1] if payload_len > 1 else None
                self._sub_revision = payload[2] if payload_len > 2 else None
            if self._is_action(ALL_INFO):
                self._board_id = payload[0] if payload_len > 0 else None
                self._num_ids = payload[1] if payload_len > 1 else None
                self._model = payload[2] if payload_len > 2 else None
                self._dc_volts = payload[3] / 10.0 if payload_len > 3 else None
        else:
            self._action_byte = action.bits if action else 0
            self.tmcc_id = int(data) if data else 0
            self._ident = ident

    def _is_action(self, enums: List[T]) -> bool:
        return self.action in enums

    def _is_command(self, enums: List[PdiCommand]) -> bool:
        for enum in enums:
            if enum == self.pdi_command:
                return True
        return False

    @property
    def is_lcs(self) -> bool:
        return True

    @property
    def is_error(self) -> bool:
        return self._error

    @property
    def error(self) -> str:
        if self.is_error and self._error_code in ERROR_CODE_MAP:
            return ERROR_CODE_MAP[self._error_code]
        else:
            return "None"

    @property
    def board_id(self) -> int | None:
        return self._board_id

    @property
    def num_ids(self) -> int | None:
        return self._num_ids

    @property
    def model(self) -> int | None:
        return self._model

    @property
    def uart0(self) -> int | None:
        return self._uart0

    @property
    def uart1(self) -> int | None:
        return self._uart1

    @property
    def dc_volts(self) -> float | None:
        return self._dc_volts

    @property
    def base_type(self) -> str | None:
        if self._base_type in BASE_TYPE_MAP:
            return BASE_TYPE_MAP[self._base_type]
        return "NA"

    @staticmethod
    def uart_mode(uart: int) -> str | None:
        if uart in UART_MAP:
            return UART_MAP[uart]
        return "NA"

    def __repr__(self) -> str:
        if self.payload is not None:
            payload = " " + self.payload
        elif self._data is not None:
            payload = f" (0x{self._data.hex()})" if self._data else ""
        else:
            payload = ""

        return f"[PDI {self._pdi_command.name} ID: {self.tmcc_id} {self.action.name}{payload}]"

    @property
    def ident(self) -> int:
        return self._ident

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return f"Error: {self.error}"
        elif self._is_action(ALL_STATUS) and self.pdi_command.name.endswith("_RX"):
            return (
                f"Board ID: {self.board_id} Num IDs: {self.num_ids} Model: {self.model} DC Volts: {self.dc_volts} "
                f"Base: {self.base_type} UART0: {self.uart_mode(self.uart0)} UART1: {self.uart_mode(self.uart1)} "
                f"\n({self.packet})"
            )
        elif self._is_action(ALL_INFO) and self.pdi_command.name.endswith("_RX"):
            return (
                f"Board ID: {self.board_id} Num IDs: {self.num_ids} Model: {self.model} "
                f"DC Volts: {self.dc_volts} ({self.packet})"
            )
        elif self._is_action(ALL_FIRMWARE) and self.pdi_command.name.endswith("_RX"):
            return f"Firmware {self._version}.{self._revision}.{self._sub_revision}"
        elif self._is_action(ALL_IDENTIFY):
            if self._is_command(ALL_SETs):
                return f"Ident: {self.ident} ({self.packet})"
        return super().payload

    @property
    def action(self) -> T:
        return self._action


class Ser2Req(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.SER2_GET,
        action: Ser2Action = Ser2Action.CONFIG,
        ident: int | None = None,
        error: bool = False,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        if isinstance(data, bytes):
            self._action = Ser2Action(self._action_byte)



from __future__ import annotations

from .sequence_constants import SequenceCommandEnum
from .sequence_req import SequenceReq
from ..constants import CommandScope, DEFAULT_ADDRESS
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum, tmcc2_speed_to_rpm


class AbsoluteSpeedRpm(SequenceReq):
    def __init__(
        self,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        data: int = 0,
    ) -> None:
        super().__init__(SequenceCommandEnum.ABSOLUTE_SPEED_RPM, address, scope)
        self._scope = scope
        self._data = data
        self._state = None
        self.add(TMCC2EngineCommandEnum.ABSOLUTE_SPEED, data=data, scope=CommandScope.ENGINE)
        rpm = tmcc2_speed_to_rpm(data)
        self.add(TMCC2EngineCommandEnum.DIESEL_RPM, data=rpm, scope=CommandScope.ENGINE)

    @property
    def scope(self) -> CommandScope | None:
        return self._scope

    @scope.setter
    def scope(self, new_scope: CommandScope) -> None:
        if new_scope != self._scope:
            # can only change scope for Engine and Train commands, and then, just from the one to the other
            if self._scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                if new_scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                    self._scope = new_scope
                    self._apply_scope()
                    return
            raise AttributeError(f"Scope {new_scope} not supported for {self}")

    def _apply_data(self, new_data: int = None) -> int:
        from ...db.component_state_store import ComponentStateStore

        state = ComponentStateStore.get_state(self.scope, self.address, create=False)
        if state:
            new_speed = min(state.speed_max, self.data)
            self._data = new_speed
        else:
            new_speed = self.data

        for req_wrapper in self._requests:
            req = req_wrapper.request
            if req.command == TMCC2EngineCommandEnum.DIESEL_RPM:
                req.data = tmcc2_speed_to_rpm(new_speed)
            elif req.command == TMCC2EngineCommandEnum.ABSOLUTE_SPEED:
                req.data = new_speed
        return 0


SequenceCommandEnum.ABSOLUTE_SPEED_RPM.value.register_cmd_class(AbsoluteSpeedRpm)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from gpiozero import PingServer

from ..comm.comm_buffer import CommBuffer
from ..utils.ip_tools import find_base_address
from .gpio_device import GpioDevice, P


class BaseWatcher(GpioDevice):
    """
    BaseWatcher class monitors connectivity to a Lionel Base 3 and controls GPIO devices accordingly.

    This class is designed to interface with a server and orchestrates the behavior of active
    and inactive LEDs based on the server's availability. The class determines the server's
    address if not explicitly provided, uses a PingServer to monitor server activity, and
    associates GPIO-based LEDs to visually reflect the activity status of the server.

    Attributes:
        active_led (Optional[GpioLED]): LED to indicate active server status.
        inactive_led (Optional[GpioLED]): LED to indicate inactive server status.

    Parameters:
        server (str, optional): The Base 3's address to monitor. If None,
            it attempts to find the address automatically.
        active_pin (P, optional): GPIO pin for the active state LED. None if
            not specified.
        inactive_pin (P, optional): GPIO pin for the inactive state LED. None if
            not specified.
        cathode (bool): If True, sets up the LEDs in cathode configuration.
            Defaults to True.
        delay (float): Delay in seconds between Base 3 status checks.
            Defaults to 10.

    Raises:
        ValueError: If the Base 3's address cannot be determined, and no `server`
            parameter is provided.
    """

    def __init__(
        self,
        server: str = None,
        active_pin: P = None,
        inactive_pin: P = None,
        cathode: bool = True,
        delay: float = 10,
    ) -> None:
        # if server isn't specified, try to figure it out
        if server is None:
            server = CommBuffer().base3_address
        if server is None:
            print("Looking for Base 3 on local network...")
            server = find_base_address()
            if server is None:
                raise ValueError("Could not determine base address")

        # set up a ping server, treat it as a device
        ping_server = PingServer(server, event_delay=delay)
        self.cache_device(ping_server)

        # set up active led, if any
        if active_pin:
            self.active_led = self.make_led(active_pin, cathode=cathode)
            self.active_led.value = 1 if ping_server.is_active else 0
            self.cache_device(self.active_led)
        else:
            self.active_led = None

        if inactive_pin:
            self.inactive_led = self.make_led(inactive_pin, cathode=cathode)
            self.inactive_led.value = 0 if ping_server.is_active else 1
            self.cache_device(self.inactive_led)
        else:
            self.inactive_led = None

        # set ping server state change actions
        if self.active_led and self.inactive_led:
            # we have to toggle the LEDs; we need a custom function
            def on_active() -> None:
                self.active_led.on()
                self.inactive_led.off()

            ping_server.when_activated = on_active

            def on_inactive() -> None:
                self.active_led.off()
                self.inactive_led.on()

            ping_server.when_deactivated = on_inactive
        elif self.active_led:
            ping_server.when_activated = self.active_led.on
            ping_server.when_deactivated = self.active_led.off
        elif self.inactive_led:
            ping_server.when_activated = self.inactive_led.off
            ping_server.when_deactivated = self.inactive_led.on



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import getpass
import ipaddress
import os
import platform
import pwd
import shutil
import subprocess
import sys
import tempfile
from argparse import ArgumentParser
from pathlib import Path
from typing import Dict

from ..utils.path_utils import find_dir, find_file
from .pytrain import DEFAULT_BUTTONS_FILE


class MakeService:
    def __init__(self, cmd_line: list[str] = None) -> None:
        from .. import is_package

        self._user = getpass.getuser()
        self._home = Path.home()
        self._cwd = Path.cwd()
        self._prog = "make_service" if is_package() else "make_service.py"
        if cmd_line:
            args = self.command_line_parser().parse_args(cmd_line)
        else:
            args = self.command_line_parser().parse_args()
        self._args = args

        # verify username
        self._user = args.user
        if self._user is None:
            print("\nA valid Raspberry Pi username is required")
            return
        elif self.validate_username(self._user) is False:
            print(f"\nUser '{self._user}' does not exist on this system. Exiting.")
            return

        if args.remove is True:
            self.deactivate_and_remove_services()
            return

        # verify template directory exists
        self._template_dir = find_dir("installation", (".", "../", "src"))
        if self._template_dir is None:
            print("\nUnable to locate directory with installation templates. Exiting")
            return

        # verify we can activate the virtual environment
        self._activate_cmd = find_file("activate", (".", "../"))
        if self._activate_cmd is None:
            print("\nUnable to locate virtual environment 'activate' command. Exiting")
            return

        # if server, verify a base 3 and/or ser2 is specified
        self._ser2 = args.ser2 is True
        if args.base is None:
            self._base_ip = None
        else:
            self._base_ip = args.base if args.base and args.base != "search" else "search"
            if self._base_ip != "search":
                if self.is_valid_ip(self._base_ip) is False:
                    print(f"\nInvalid IP address '{self._base_ip}'. Exiting")
                    return
            else:
                self._base_ip = ""  # an empty value causes PyTrain to search for the base
        if args.mode == "server" and args.base is None and args.ser2 is False:
            print("\nA Lionel Base IP address or Ser2 is required when configuring as a server. Exiting")
            return

        # verify client args
        if args.mode == "client":
            if self._base_ip is not None:
                print("\nA Lionel Base IP address is not required when configuring as a client. Continuing")
            if args.ser2 is True:
                print("\nA Ser2 is not required when configuring as a client. Continuing")

        # verify buttons file exists
        if args.buttons_file:
            if os.path.isfile(args.buttons_file) is False:
                print(f"\nButton definitions file '{args.buttons_file}' not found. Continuing")

        self._exe = "pytrain" if is_package() else "cli/pytrain.py"
        self._echo = args.echo is True
        self._cmd_line = self.command_line
        self._config = {
            "___ACTIVATE___": str(self._activate_cmd),
            "___BUTTONS___": self._args.buttons_file if self._args.buttons_file else "",
            "___CLIENT___": "-client" if self.is_client else "",
            "___ECHO___": " -echo" if self._echo is True else "",
            "___HOME___": str(self._home),
            "___LCSSER2___": " -ser2" if self._ser2 is True else "",
            "___LIONELBASE___": f"-base {self._base_ip}" if self._base_ip is not None else "",
            "___MODE___": "Server" if self.is_server else "Client",
            "___PYTRAINHOME___": str(self._cwd),
            "___PYTRAIN___": str(self._exe),
            "___USER___": self._user,
        }
        self._start_service = args.start is True
        if self.confirm_environment():
            path = self.make_shell_script()
            if path:
                self._config["___SHELL_SCRIPT___"] = str(path)
                self.install_service()
        else:
            print("\nRe-run this script with the -h option for help")

    def make_shell_script(self) -> Path | None:
        template = find_file("pytrain.bash.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate shell script template. Exiting")
            return None
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        path = Path(self._home, "pytrain_server.bash" if self.is_server else "pytrain_client.bash")
        # write the shell script file
        if path.exists():
            shutil.copy2(path, path.with_suffix(".bak"))
        with open(path, "w") as f:
            f.write(template_data)
        os.chmod(path, 0o755)
        print(f"\n{path} created")
        return path

    def install_service(self) -> str | None:
        from .. import PROGRAM_NAME

        if platform.system().lower() != "linux":
            print(f"\nPlease run {self._prog} from a Raspberry Pi. Exiting")
            return None
        template = find_file("pytrain.service.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate service definition template. Exiting")
            return None
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        tmp = tempfile.NamedTemporaryFile()
        with open(tmp.name, "w") as f:
            f.write(template_data)
        service = "pytrain_server.service" if self.is_server else "pytrain_client.service"
        result = subprocess.run(
            f"sudo cp -f {tmp.name} /etc/systemd/system/{service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error creating /etc/systemd/system/{service}: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            f"sudo chmod 644 /etc/systemd/system/{service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error changing mode of /etc/systemd/system/{service}: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            "sudo systemctl daemon-reload".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error reloading system daemons: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            f"sudo systemctl enable {service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error enabling {PROGRAM_NAME} service: {result.stderr} Exiting")
            return None
        if self._start_service:
            subprocess.run(
                f"sudo systemctl restart {service}".split(),
            )
            print(f"\n{PROGRAM_NAME} service started...")
        return service

    @property
    def is_client(self) -> bool:
        return self._args.mode == "client"

    @property
    def is_server(self) -> bool:
        return self._args.mode == "server"

    @property
    def config(self) -> Dict[str, str]:
        return self._config

    @property
    def pytrain_path(self) -> str:
        return f"{self._cwd}/{self._exe}"

    @property
    def command_line(self) -> str | None:
        cmd_line = f"{self._exe} -headless"
        if self._args.mode == "client":
            cmd_line += " -client"
        else:
            if self._base_ip:
                ip = self._base_ip
                ip = f" {ip}" if ip != "search" else ""
                cmd_line += f" -base{ip}"
            if self._args.ser2 is True:
                cmd_line += " -ser2"
        if self._echo is True:
            cmd_line += " -echo"
        if self._args.buttons_file:
            cmd_line += f" -buttons {self._args.buttons_file}"
        return cmd_line

    def confirm_environment(self) -> bool:
        from .. import PROGRAM_NAME

        print(f"\nInstalling {PROGRAM_NAME} as a systemd service with these settings:")
        print(f"  Mode: {'Client' if self._args.mode == 'client' else 'Server'}")
        if self._args.mode == "server":
            print(f"  Lionel Base IP addresses: {self._base_ip}")
            print(f"  Use Ser2: {'Yes' if self._args.ser2 is True else 'No'}")
        print(f"  Button definitions file: {self._args.buttons_file if self._args.buttons_file else 'None'}")
        print(f"  Run as user: {self._user}")
        print(f"  User '{self._user} Home: {self._home}")
        print(f"  Echo TMCC/Legacy/Pdi commands to log file: {'Yes' if self._echo is True else 'No'}")
        print(f"  System type: {platform.system()}")
        print(f"  Virtual environment activation command: {self._activate_cmd}")
        print(f"  {PROGRAM_NAME} Exe: {self._exe}")
        print(f"  {PROGRAM_NAME} Home: {self._cwd}")
        print(f"  {PROGRAM_NAME} Command Line: {self._cmd_line}")
        return self.confirm("\nConfirm? [y/n] ")

    @staticmethod
    def confirm(msg: str = None) -> bool:
        msg = msg if msg else "Continue? [y/n] "
        answer = input(msg)
        return True if answer.lower() in ["y", "yes"] else False

    @staticmethod
    def is_valid_ip(ip: str) -> bool:
        try:
            ipaddress.ip_address(ip)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_service_present(service: str) -> bool:
        cmd = f"sudo systemctl status {service}.service"
        result = subprocess.run(cmd.split(), capture_output=True)
        if result.returncode == 4 or os.path.exists(f"/etc/systemd/system/{service}.service") is False:
            return False
        else:
            return True

    @staticmethod
    def deactivate_service(service: str) -> None:
        subprocess.run(f"sudo systemctl stop {service}.service".split())
        subprocess.run(f"sudo systemctl disable {service}.service".split())
        subprocess.run(f"sudo rm -fr /etc/systemd/system/{service}.service".split())
        subprocess.run("sudo systemctl daemon-reload".split())
        subprocess.run("sudo systemctl reset-failed".split())

    @property
    def is_server_present(self) -> bool:
        return self.is_service_present("pytrain_server")

    @property
    def is_client_present(self) -> bool:
        return self.is_service_present("pytrain_client")

    def deactivate_and_remove_services(self):
        from .. import PROGRAM_NAME

        if platform.system().lower() != "linux":
            print(f"\nPlease run {self._prog} from a Raspberry Pi. Exiting")
            return

        if self.is_server_present:
            mode = "Server"
        elif self.is_client_present:
            mode = "Client"
        else:
            print(f"\nNo {PROGRAM_NAME} server or client is currently running. Exiting")
            return
        if self.confirm(f"Are you sure you want to deactivate and remove {PROGRAM_NAME} {mode}?") is True:
            path = Path(self._home, "pytrain_server.bash" if mode == "Server" else "pytrain_client.bash")
            if path.exists():
                print(f"\nRemoving {path}...")
                path.unlink(missing_ok=True)
            print(f"\nDeactivating {PROGRAM_NAME} {mode} service...")
            self.deactivate_service("pytrain_server" if mode == "Server" else "pytrain_client")

    def command_line_parser(self) -> ArgumentParser:
        from .. import PROGRAM_NAME, get_version

        parser = ArgumentParser(
            prog=self._prog,
            description=f"Launch {PROGRAM_NAME} as a systemd service when your Raspberry Pi is powered on",
        )
        mode_group = parser.add_mutually_exclusive_group(required=True)
        mode_group.add_argument(
            "-client",
            action="store_const",
            const="client",
            dest="mode",
            help=f"Configure this node as a {PROGRAM_NAME} client",
        )
        mode_group.add_argument(
            "-server",
            action="store_const",
            const="server",
            dest="mode",
            help=f"Configure this node as a {PROGRAM_NAME} server",
        )
        mode_group.add_argument(
            "-remove",
            action="store_true",
            help="Deactivate and remove any existing {PROGRAM_NAME} service",
        )
        server_opts = parser.add_argument_group("Server options")
        server_opts.add_argument(
            "-base",
            nargs="?",
            default=None,
            const="search",
            help="IP address of Lionel Base 3 or LCS Wi-Fi module",
        )
        server_opts.add_argument(
            "-ser2",
            action="store_true",
            help="Send or receive TMCC commands from an LCS Ser2",
        )
        misc_opts = parser.add_argument_group("Miscellaneous options")
        misc_opts.add_argument(
            "-buttons_file",
            nargs="?",
            default=None,
            const=DEFAULT_BUTTONS_FILE,
            help=f"Button definitions file, loaded when {PROGRAM_NAME} starts",
        )
        misc_opts.add_argument(
            "-echo",
            action="store_true",
            help="Echo received TMCC/Legacy/Pdi commands to log file",
        )
        misc_opts.add_argument(
            "-start",
            action="store_true",
            help=f"Start {PROGRAM_NAME} Client/Server now (otherwise, it starts on reboot)",
        )
        misc_opts.add_argument(
            "-user",
            action="store",
            default=self._user,
            help=f"Raspberry Pi user to run {PROGRAM_NAME} as (default: {self._user})",
        )
        misc_opts.add_argument(
            "-version",
            action="version",
            version=f"{self.__class__.__name__} {get_version()}",
            help="Show version and exit",
        )
        return parser

    @staticmethod
    def validate_username(user: str) -> bool:
        try:
            pwd.getpwnam(user)
            return True
        except KeyError:
            return False


def main(args: list[str] | None = None) -> int:
    if args is None:
        args = sys.argv[1:]
    try:
        MakeService(args)
        return 0
    except Exception as e:
        # Output anything else nicely formatted on stderr and exit code 1
        sys.exit(f"{__file__}: error: {e}\n")



from __future__ import annotations

import sys
from abc import ABC, ABCMeta, abstractmethod
from time import sleep
from typing import Tuple, TypeVar, List

from ..utils.validations import Validations

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from .constants import PDI_SOP, PDI_EOP, PDI_STF, CommonAction, PdiAction, PdiCommand
from ..protocol.constants import CommandScope, MINIMUM_DURATION_INTERVAL_MSEC, DEFAULT_DURATION_INTERVAL_MSEC

T = TypeVar("T", bound=PdiAction)

LIONEL_ENGINE_RECORD_LENGTH: int = 0xC0
SCOPE_TO_RECORD_LENGTH = {
    CommandScope.ENGINE: LIONEL_ENGINE_RECORD_LENGTH,
    CommandScope.TRAIN: LIONEL_ENGINE_RECORD_LENGTH,
    CommandScope.ACC: LIONEL_ENGINE_RECORD_LENGTH,
    CommandScope.SWITCH: 0x40,
    CommandScope.ROUTE: 0x80,
}


# noinspection GrazieInspection
class PdiReq(ABC):
    __metaclass__ = ABCMeta

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        # throws an exception if we can't dereference
        pdi_cmd = PdiCommand(data[1])
        try:
            from .pdi_device import PdiDevice

            dev = PdiDevice.from_pdi_command(pdi_cmd)
            return dev.build_req(data)
        except AttributeError as ae:
            raise ae
        except ValueError as ve:
            raise ve

    @classmethod
    def scope_record_length(cls, scope: CommandScope) -> int:
        return SCOPE_TO_RECORD_LENGTH.get(scope, None)

    @abstractmethod
    def __init__(self, data: bytes | None, pdi_command: PdiCommand = None) -> None:
        super().__init__()
        # default scope is system; override as needed in child classes
        # also change as needed when sending command updates to state handler
        self._scope = CommandScope.SYSTEM
        self._tmcc_id = 0
        if isinstance(data, bytes):
            # first byte should be SOP, last byte should be EOP, if not, raise exception
            if data[0] != PDI_SOP or data[-1] != PDI_EOP:
                raise ValueError(f"Invalid PDI Request: {data}")
            # process the data to remove SOP, EOP, Checksum, and Stuff Bytes, if any
            self._original = data
            recv_checksum = data[-2]
            self._data, check_sum = self._calculate_checksum(data[1:-2], False)
            if recv_checksum != int.from_bytes(check_sum, byteorder="big"):
                raise ValueError(
                    f"Invalid PDI Request: 0x{data.hex()} {hex(recv_checksum)} != {check_sum.hex()} [BAD CHECKSUM]"
                )
            self._pdi_command: PdiCommand = PdiCommand(data[1])
        else:
            if pdi_command is None or not isinstance(pdi_command, PdiCommand):
                raise ValueError(f"Invalid PDI Request: {pdi_command}")
            self._pdi_command = pdi_command
            self._data = self._original = None
        # TODO: block setter
        from .pdi_device import PdiDevice

        self.pdi_device: PdiDevice = PdiDevice.from_pdi_command(self.pdi_command)

    def __repr__(self) -> str:
        return f"[PDI {self._pdi_command.friendly} {self.payload}]"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    @staticmethod
    def _calculate_checksum(data: bytes, add_stf=True) -> Tuple[bytes, bytes]:
        """
        Used to calculate checksums on packets we send as well as to strip Stuff bytes
        from packets we receive
        """
        byte_stream = bytes()
        check_sum = 0
        last_byte = None
        for b in data:
            check_sum += b
            if b in {PDI_SOP, PDI_STF, PDI_EOP}:
                if add_stf is True:
                    # add the stuff byte to the output and account for it in the check sum
                    check_sum += PDI_STF
                    byte_stream += PDI_STF.to_bytes(1, byteorder="big")
                elif b in {PDI_SOP, PDI_EOP}:
                    # we are parsing a received packet; strip stuff byte
                    pass  # we want this byte added to the output stream
                else:
                    # this must be a stuff byte, and we are receiving; strip it
                    # unless the previous byte was also a stuff byte
                    if last_byte != PDI_STF:
                        last_byte = b
                        continue
            byte_stream += b.to_bytes(1, byteorder="big")
            last_byte = b
        # do checksum calculation on buffer
        byte_sum = check_sum
        check_sum = 0xFF & (0 - check_sum)
        if check_sum in {PDI_SOP, PDI_STF, PDI_EOP}:
            if add_stf is True:
                byte_stream += PDI_STF.to_bytes(1, byteorder="big")
            byte_sum += PDI_STF
            check_sum = 0xFF & (0 - byte_sum)
        return byte_stream, check_sum.to_bytes(1, byteorder="big")

    @property
    def pdi_command(self) -> PdiCommand:
        return self._pdi_command

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @scope.setter
    def scope(self, scope: CommandScope) -> None:
        if scope is None or not isinstance(scope, CommandScope):
            raise ValueError(f"Invalid CommandScope: {scope}")
        self._scope = scope

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @tmcc_id.setter
    def tmcc_id(self, tmcc_id: int) -> None:
        self._tmcc_id = tmcc_id

    @property
    def address(self) -> int:
        # for compatibility with state management system
        return self.tmcc_id

    @property
    def command(self) -> PdiCommand:
        # for compatibility with state management system
        return self.pdi_command

    @property
    def action(self) -> T | None:
        return None

    @property
    def base_record_length(self) -> int:
        """
        Return the record length of the PdiCommand.BASE_MEMORY or
        D4_ENGINE/D4_TRAIN record in the Base 3 based on Scope.
        """
        return SCOPE_TO_RECORD_LENGTH.get(self.scope, None)

    @property
    def as_bytes(self) -> bytes:
        # if this request was constructed from a byte stream, return
        # the original stream if we call as_bytes
        if self._original:
            return self._original
        """
        Default implementation, should override in more complex requests
        """
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        if self.action is None:
            byte_str += CommonAction.CONFIG.as_bytes
        else:
            byte_str += self.action.as_bytes
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @property
    def checksum(self) -> bytes:
        check_sum = 0
        for b in self._data:
            if b in {PDI_SOP, PDI_EOP}:
                continue
            check_sum += b
        check_sum = 0xFF & (0 - check_sum)
        return check_sum.to_bytes(1)

    @property
    def is_ping(self) -> bool:
        return self._pdi_command == PdiCommand.PING

    @property
    def is_ack(self) -> bool:
        return False

    @property
    def is_tmcc(self) -> bool:
        return self._pdi_command.is_tmcc

    @property
    def is_lcs(self) -> bool:
        return False

    @property
    def is_tmcc_rx(self) -> bool:
        return False

    @property
    def is_filtered(self) -> bool:
        return False

    @staticmethod
    def decode_text(data: bytes) -> str | None:
        name = ""
        num_ffs = 0
        for b in data:
            if b == 0:
                break
            elif b == 0xFF:
                num_ffs += 1
            else:
                name += chr(b)
        if num_ffs == len(data):
            return None
        return name

    @staticmethod
    def encode_text(text: str, field_len: int) -> bytes | None:
        if text is None:
            return b"\x00" * field_len
        else:
            return bytes(text, "ascii").ljust(field_len, b"\x00")

    @property
    def payload(self) -> str:
        return f"({self.packet})"

    @property
    def packet(self) -> str:
        if self._data is None:
            return "0x" + self.as_bytes.hex(" ").upper()
        else:
            return "0x" + self._data.hex(" ").upper()

    @property
    def as_key(self):
        return self.tmcc_id, self.pdi_command, self.action, self.scope

    def send(
        self,
        repeat: int = 1,
        delay: float = 0.0,
        duration: float = 0.0,
        interval: int = None,
        baudrate: int = None,
        port: str | int = None,
        server: str = None,
    ) -> None:
        """
        Send PDI command bytes to TMCC Buffer for dispatch; the TMCC buffer
        will dispatch to the appropriate sender (tmcc or pdi)
        """
        from ..comm.comm_buffer import CommBuffer

        repeat = Validations.validate_int(repeat, min_value=1, label="repeat")
        delay = Validations.validate_float(delay, min_value=0, label="delay")
        duration = Validations.validate_float(duration, min_value=0, label="duration", allow_none=True)
        interval = Validations.validate_int(
            interval,
            min_value=MINIMUM_DURATION_INTERVAL_MSEC,
            label="interval",
            allow_none=True,
        )

        buffer = CommBuffer.build(baudrate=baudrate, port=port, server=server)
        for rep_no in range(repeat):
            buffer.enqueue_command(self.as_bytes, delay=delay)
            if duration > 0:
                interval = interval if interval else DEFAULT_DURATION_INTERVAL_MSEC
                # convert duration into milliseconds, then queue a command to fire
                # every 100 msec for the duration
                pause = 0
                for d in range(interval, int(round(duration * 1000)), interval):
                    buffer.enqueue_command(self.as_bytes, delay + (d / 1000.0) - pause)
                    pause += 0.002
                    sleep(0.002)


class TmccReq(PdiReq):
    from ..protocol.command_req import CommandReq

    @classmethod
    def as_packets(cls, tmcc_cmd: CommandReq) -> List[bytes]:
        """
        Used to decompose multibyte TMCC commands into 3-byte packets, sending each
        as a PdiCommand.TMCC_TX packet

        Also support 4-digit engines and trains using PdiCommand.TMCC4_RX
        """
        byte_str = tmcc_cmd.as_bytes
        if tmcc_cmd.address > 99:
            pdi_cmd = PdiCommand.TMCC4_TX
            step_size = 7
        else:
            pdi_cmd = PdiCommand.TMCC_TX
            step_size = 3
        packets = []
        for i in range(0, len(byte_str), step_size):
            packet = pdi_cmd.to_bytes(1, byteorder="big") + byte_str[i : i + step_size]
            packet, checksum = cls._calculate_checksum(packet)
            packet = PDI_SOP.to_bytes(1, byteorder="big") + packet
            packet += checksum
            packet += PDI_EOP.to_bytes(1, byteorder="big")
            packets.append(packet)
        return packets

    def __init__(self, data: bytes | CommandReq, pdi_command: PdiCommand = None):
        super().__init__(data, pdi_command=pdi_command)
        from ..protocol.command_req import CommandReq

        if isinstance(data, CommandReq):
            if self.pdi_command.is_tmcc is False:
                raise ValueError(f"Invalid PDI TMCC Request: {self.pdi_command}")
            self._tmcc_command: CommandReq = data
        else:
            pdi_command = PdiCommand(data[1])
            if pdi_command in {PdiCommand.TMCC4_TX, PdiCommand.TMCC4_RX}:
                self._tmcc_command: CommandReq = CommandReq.from_bytes(
                    self._data[1:],
                    from_tmcc_rx=True if pdi_command == PdiCommand.TMCC4_RX else False,
                    is_tmcc4=True,
                )
            elif pdi_command in {PdiCommand.TMCC_TX, PdiCommand.TMCC_RX}:
                self._tmcc_command: CommandReq = CommandReq.from_bytes(
                    self._data[1:], from_tmcc_rx=True if pdi_command == PdiCommand.TMCC_RX else False
                )
            else:
                raise ValueError(f"Invalid PDI TMCC Request: {data}")

    def __repr__(self) -> str:
        data = f" (0x{self._data.hex()})" if self._data else ""
        return f"[PDI {self._pdi_command.friendly} {self.tmcc_command}{data}]"

    @property
    def tmcc_command(self) -> CommandReq:
        return self._tmcc_command

    @property
    def is_tmcc_rx(self) -> bool:
        return self._pdi_command in {PdiCommand.TMCC_RX, PdiCommand.TMCC4_RX}

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes + self.tmcc_command.as_bytes
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @property
    def scope(self) -> CommandScope:
        return CommandScope.SYSTEM

    @property
    def is_tmcc4(self) -> bool:
        return self.address > 99


class AllReq(PdiReq):
    def __init__(self, data: bytes = None, pdi_command: PdiCommand = PdiCommand.ALL_GET) -> None:
        super().__init__(data, pdi_command)

    @property
    def payload(self) -> str | None:
        return None

    @property
    def scope(self) -> CommandScope:
        return CommandScope.SYSTEM


class PingReq(PdiReq):
    def __init__(self, data: bytes | None = None):
        super().__init__(data, PdiCommand.PING)
        if data is not None:
            if PdiCommand(data[1]).is_ping is False:
                raise ValueError(f"Invalid PDI Ping Request: {data}")

    @property
    def as_bytes(self) -> bytes:
        byte_str = self.pdi_command.as_bytes
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @property
    def scope(self) -> CommandScope:
        return CommandScope.SYSTEM

    def __repr__(self) -> str:
        return f"[PDI {self._pdi_command.friendly}]"



from .tmcc1_command import TMCC1Command
from .tmcc1_constants import TMCC1RouteCommandEnum

from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT


class RouteCmd(TMCC1Command):
    def __init__(
        self, route: int, baudrate: int = DEFAULT_BAUDRATE, port: str = DEFAULT_PORT, server: str = None
    ) -> None:
        if route < 1 or route > 31:
            raise ValueError("TMCC1 Route must be between 1 and 31")
        req = CommandReq(TMCC1RouteCommandEnum.FIRE, route)
        super().__init__(TMCC1RouteCommandEnum.FIRE, req, route, 0, None, baudrate, port, server)
        self._command = self._build_command()



import time
from collections import defaultdict
from unittest import mock

# noinspection PyPackageRequirements
import pytest

# noinspection PyProtectedMember
from src.pytrain.comm.command_listener import CommandDispatcher, CommandListener, Message
from src.pytrain.db.component_state import (
    SCOPE_TO_STATE_MAP,
    ComponentState,
    ComponentStateDict,
    SwitchState,
    SystemStateDict,
)
from src.pytrain import EngineState, TrainState, AccessoryState
from src.pytrain.protocol.command_req import CommandReq
from src.pytrain.protocol.constants import BROADCAST_ADDRESS, CommandScope
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum as Acc
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1HaltCommandEnum
from src.pytrain.protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum as Switch
from src.pytrain.protocol.tmcc2.tmcc2_constants import TMCC2RouteCommandEnum

from ..test_base import TestBase


@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmpdir) -> None:
    """
    Fixture to execute asserts before and after a test is run
    """
    # Setup: fill with any logic you want

    # noinspection PyTypeChecker
    yield  # this is where the testing happens

    # Teardown : fill with any logic you want
    if CommandListener.is_built():
        CommandListener().shutdown()
    assert CommandListener.is_built() is False

    if CommandDispatcher.is_built():
        CommandDispatcher().shutdown()
    assert CommandDispatcher.is_built() is False


# noinspection PyMethodMayBeStatic
class TestComponentState(TestBase):
    state: dict[CommandScope, ComponentStateDict] = SystemStateDict()

    def __call__(self, message: Message) -> None:
        if message:
            if message.is_halt:  # send to all known devices
                print("*** Received HALT Command ***")
            elif message.is_system_halt:  # send to all known engines and trains
                print("*** Received SYSTEM HALT Command ***")
            elif message.scope in SCOPE_TO_STATE_MAP:
                if message.address == BROADCAST_ADDRESS:  # broadcast address
                    print(f"*** {message.scope.name} Broadcast Address ***")
                else:  # update the device state (identified by scope/address)
                    self.state[message.scope][message.address].update(message)

    def test_component_state(self) -> None:
        with pytest.raises(TypeError):
            # noinspection PyTypeCher
            ComponentState(None)

    def test_switch_state(self) -> None:
        """
        Test that switch state is correctly initialized and that behavior
        implemented in the parent ComponentState class is as expected
        """
        ss = SwitchState()
        assert ss is not None
        assert ss.scope == CommandScope.SWITCH
        assert ss.address is None
        assert ss.last_command is None
        assert ss.last_updated is None
        assert ss.state is None
        assert ss.is_known is False
        assert ss.is_through is False
        assert ss.is_out is False

        # assert construction with scope Switch succeeds
        ss = SwitchState(CommandScope.SWITCH)
        assert ss is not None
        assert ss.scope == CommandScope.SWITCH

        # assert construction with any other scope fails
        for scope in CommandScope:
            if scope == CommandScope.SWITCH:
                continue
            with pytest.raises(ValueError):
                # noinspection PyTypeChecker
                SwitchState(scope)

        # assert calling update with a valid request succeeds
        set_req = CommandReq(Switch.SET_ADDRESS, 1)
        assert set_req is not None
        assert set_req.scope == CommandScope.SWITCH
        assert set_req.address == 1

        # now call update
        ss.update(set_req)
        assert ss.last_command == set_req
        assert ss.last_updated is not None
        last_updated = ss.last_updated
        assert ss.address == 1
        # state should still be unknown
        assert ss.state is None
        assert ss.is_known is False
        assert ss.is_through is False
        assert ss.is_out is False

        # create a "switch thrown to out" request
        out_req = CommandReq(Switch.OUT, 1)
        ss.update(out_req)
        assert ss.last_command == out_req
        assert ss.last_updated is not None
        assert ss.last_updated >= last_updated
        last_updated = ss.last_updated
        assert ss.address == 1
        assert ss.state == Switch.OUT
        assert ss.is_known is True
        assert ss.is_out is True
        assert ss.is_through is False

        # create a "switch thrown to through" request
        thru_req = CommandReq(Switch.THRU, 1)
        ss.update(thru_req)
        assert ss.last_command == thru_req
        assert ss.last_updated is not None
        assert ss.last_updated >= last_updated
        last_updated = ss.last_updated
        assert ss.address == 1
        assert ss.state == Switch.THRU
        assert ss.is_known is True
        assert ss.is_out is False
        assert ss.is_through is True

        # update with another "set address" request, state should be unchanged
        ss.update(set_req)
        assert ss.last_command == set_req
        assert ss.last_updated is not None
        assert ss.last_updated >= last_updated
        last_updated = ss.last_updated
        assert ss.address == 1
        assert ss.state == Switch.THRU
        assert ss.is_known is True
        assert ss.is_out is False
        assert ss.is_through is True

        # throw switch out and check state is as expected
        ss.update(out_req)
        assert ss.last_command == out_req
        assert ss.last_updated is not None
        assert ss.last_updated >= last_updated
        assert ss.address == 1
        assert ss.state == Switch.OUT
        assert ss.is_known is True
        assert ss.is_out is True
        assert ss.is_through is False

        # verify address can not be modified once set
        thru_req = CommandReq(Switch.THRU, 2)
        with pytest.raises(AttributeError, match="Switch #1 received update for Switch #2, ignoring"):
            ss.update(thru_req)

        # verify we can't send update for some other object
        route_req = CommandReq(TMCC2RouteCommandEnum.FIRE, 1)
        with pytest.raises(AttributeError, match="Switch 1 received update for Route, ignoring"):
            ss.update(route_req)

        # verify we can receive a Halt command
        halt_req = CommandReq(TMCC1HaltCommandEnum.HALT, 1)
        ss.update(halt_req)
        # but that it didn't affect state
        assert ss.address == 1
        assert ss.state == Switch.OUT
        assert ss.is_known is True
        assert ss.is_out is True
        assert ss.is_through is False

    def test_accessory_state(self) -> None:
        """
        Test that AccessoryState is correctly initialized and that behavior specific
        to it is handled correctly.
        """
        acc_state = AccessoryState()
        assert acc_state is not None
        assert acc_state.scope == CommandScope.ACC
        assert acc_state.address is None
        assert acc_state.last_command is None
        assert acc_state.last_updated is None
        assert acc_state.aux_state is None
        assert acc_state.aux1_state is None
        assert acc_state.aux2_state is None
        assert acc_state.is_known is False
        assert acc_state.value is None

        # assert construction with scope Acc succeeds
        acc_state = AccessoryState(CommandScope.ACC)
        assert acc_state is not None
        assert acc_state.scope == CommandScope.ACC

        # assert construction with any other scope fails
        for scope in CommandScope:
            if scope == CommandScope.ACC:
                continue
            with pytest.raises(ValueError):
                # noinspection PyTypeChecker
                AccessoryState(scope)

        # assert calling update with a valid request succeeds
        set_req = CommandReq(Acc.SET_ADDRESS, 5)
        assert set_req is not None
        assert set_req.scope == CommandScope.ACC
        assert set_req.address == 5

        # now call update
        acc_state.update(set_req)
        assert acc_state.last_command == set_req
        assert acc_state.last_updated is not None
        last_updated = acc_state.last_updated
        assert acc_state.address == 5

        # states should still be unknown
        assert acc_state.aux_state is None
        assert acc_state.aux1_state is None
        assert acc_state.aux2_state is None
        assert acc_state.is_known is False

        # Send Aux1 Opt1 (Aux 1 btn) this should set state
        aux1_opt1_req = CommandReq(Acc.AUX1_OPT_ONE, 5)
        acc_state.update(aux1_opt1_req)
        assert acc_state.aux_state == Acc.AUX1_OPT_ONE
        assert acc_state.is_aux_on is True
        assert acc_state.is_aux_off is False
        assert acc_state.aux1_state == Acc.AUX1_ON
        assert acc_state.aux2_state is None
        assert acc_state.value is None
        assert acc_state.is_known is True
        assert acc_state.last_updated >= last_updated
        last_updated = acc_state.last_updated

        # send a numeric command
        aux1_num_req = CommandReq(Acc.NUMERIC, 5, 6)
        acc_state.update(aux1_num_req)
        assert acc_state.aux_state == Acc.AUX1_OPT_ONE
        assert acc_state.is_aux_on is True
        assert acc_state.is_aux_off is False
        assert acc_state.aux1_state == Acc.AUX1_ON
        assert acc_state.aux2_state is None
        assert acc_state.value == 6
        assert acc_state.is_known is True
        assert acc_state.last_updated >= last_updated
        last_updated = acc_state.last_updated

        # send a halt command, it should turn Aux 1 & 2 off
        halt_req = CommandReq(TMCC1HaltCommandEnum.HALT, 1)
        acc_state.update(halt_req)
        assert acc_state.aux_state == Acc.AUX2_OPT_ONE
        assert acc_state.is_aux_on is False
        assert acc_state.is_aux_off is True
        assert acc_state.aux1_state == Acc.AUX1_OFF
        assert acc_state.aux2_state == Acc.AUX2_OFF
        assert acc_state.value is None
        assert acc_state.is_known is True
        assert acc_state.last_updated >= last_updated

        # noinspection PyUnusedLocal
        def patched(arg1, arg2) -> int:
            return 2

        # turn Aux back on, then off
        with mock.patch.object(ComponentState, "time_delta", side_effect=patched):
            acc_state.update(aux1_opt1_req)
            assert acc_state.aux_state == Acc.AUX1_OPT_ONE
            assert acc_state.is_aux_on is True
            assert acc_state.is_aux_off is False
            assert acc_state.aux1_state == Acc.AUX1_ON
            assert acc_state.aux2_state is Acc.AUX2_OFF
            assert acc_state.value is None
            assert acc_state.is_known is True

        aux2_opt1_req = CommandReq(Acc.AUX2_OPT_ONE, 5)
        acc_state.update(aux2_opt1_req)
        assert acc_state.aux_state == Acc.AUX2_OPT_ONE
        assert acc_state.is_aux_on is False
        assert acc_state.is_aux_off is True
        assert acc_state.aux1_state == Acc.AUX1_ON
        assert acc_state.aux2_state is Acc.AUX2_ON
        assert acc_state.value is None
        assert acc_state.is_known is True

    def test_engine_state(self) -> None:
        """
        Test that EngineState is correctly initialized and that behavior specific
        to it is handled correctly.
        """
        eng_state = EngineState()
        assert eng_state is not None
        assert eng_state.scope == CommandScope.ENGINE
        assert eng_state.address is None
        assert eng_state.last_command is None
        assert eng_state.last_updated is None

        # assert construction with scope Acc succeeds
        eng_state = EngineState(CommandScope.ENGINE)
        assert eng_state is not None
        assert eng_state.scope == CommandScope.ENGINE

        # assert construction with any other scope fails
        for scope in CommandScope:
            if scope in [CommandScope.ENGINE, CommandScope.TRAIN]:
                continue
            with pytest.raises(ValueError):
                # noinspection PyTypeChecker
                EngineState(scope)

        # TODO: add tests once Engine State is defined

    def test_train_state(self) -> None:
        """
        Test that TrainState is correctly initialized and that behavior specific
        to it is handled correctly.
        """
        train_state = TrainState()
        assert train_state is not None
        assert train_state.scope == CommandScope.TRAIN
        assert train_state.address is None
        assert train_state.last_command is None
        assert train_state.last_updated is None

        # assert construction with scope Acc succeeds
        train_state = EngineState(CommandScope.TRAIN)
        assert train_state is not None
        assert train_state.scope == CommandScope.TRAIN

        train_state = TrainState(CommandScope.TRAIN)
        assert train_state is not None
        assert train_state.scope == CommandScope.TRAIN

        # assert construction with any other scope fails
        for scope in CommandScope:
            if scope == CommandScope.TRAIN:
                continue
            with pytest.raises(ValueError):
                # noinspection PyTypeChecker
                TrainState(scope)

        # TODO: add tests once Engine State is defined

    def test_system_state_dict(self) -> None:
        """
        SystemStateDict is used by the PyTrain cli to maintain a cache of engine, train,
        switch, and accessory state, as determined by the TMCC command stream it receives
        via publish/subscribe in the CommandListener class.
        """
        ss_dict = SystemStateDict()
        assert ss_dict is not None
        assert isinstance(ss_dict, defaultdict)
        assert isinstance(ss_dict, dict)
        assert len(ss_dict) == 0
        assert len(ss_dict.keys()) == 0
        assert len(ss_dict.values()) == 0

        # keys are instances of a subset of CommandScope
        for key in [CommandScope.ENGINE, CommandScope.TRAIN, CommandScope.SWITCH, CommandScope.ACC]:
            assert key not in ss_dict
            value = ss_dict[key]
            assert key in ss_dict
            assert value.scope == key
            assert value is not None
            assert ss_dict[key] == value  # identical keys should return same values
            assert isinstance(value, ComponentStateDict)

        # we should have an entry for each of the 4 scopes and that the entry is an empty
        # ComponentStateDict with the correct scope
        for key in [CommandScope.ENGINE, CommandScope.TRAIN, CommandScope.SWITCH, CommandScope.ACC]:
            assert key in ss_dict
            assert isinstance(ss_dict[key], ComponentStateDict)
            assert len(ss_dict[key]) == 0
            assert ss_dict[key] is not None
            assert ss_dict[key].scope == key

        # should not be able to add other keys
        with pytest.raises(KeyError, match="Invalid scope key: 123"):
            _ = ss_dict[123]

        # including CommandScopes other than the 4 above
        for key in CommandScope:
            if key in [
                CommandScope.ENGINE,
                CommandScope.TRAIN,
                CommandScope.SWITCH,
                CommandScope.ROUTE,
                CommandScope.ACC,
                CommandScope.IRDA,
                CommandScope.BASE,
                CommandScope.SYNC,
                CommandScope.BLOCK,
            ]:
                continue
            with pytest.raises(KeyError, match=f"Invalid scope key: {key}"):
                _ = ss_dict[key]

    def test_component_state_dict(self) -> None:
        """
        ComponentStateDict saves system state for engines, trains, switches, and accessories
        as determined by the TMCC command stream. CommandScope-appropriate dicts are
        constructed by SystemStateDict as needed, using the defailtdict mechanism.

        ComponentStateDict keys themselves are ints between 1 and 99 inclusive.
        Except for the command base itself, which has an ID of 0
        """
        # test that all four types of ComponentStateDicts are built
        for scope in [
            CommandScope.ENGINE,
            CommandScope.TRAIN,
            CommandScope.SWITCH,
            CommandScope.ACC,
            CommandScope.IRDA,
            CommandScope.BASE,
        ]:
            cs_dict = ComponentStateDict(scope)
            assert isinstance(cs_dict, ComponentStateDict)
            assert cs_dict.scope == scope
            value = cs_dict[1] if scope != CommandScope.BASE else cs_dict[0]
            assert value is not None
            assert isinstance(value, ComponentState)
            assert value.scope == scope

        # verify we cannot construct a dict from an invalid scope
        for key in CommandScope:
            if key in [
                CommandScope.ENGINE,
                CommandScope.TRAIN,
                CommandScope.SWITCH,
                CommandScope.ROUTE,
                CommandScope.ACC,
                CommandScope.IRDA,
                CommandScope.BASE,
                CommandScope.SYNC,
                CommandScope.BLOCK,
            ]:
                continue
            with pytest.raises(ValueError, match=f"Invalid scope: {key}"):
                ComponentStateDict(key)

        # verify dicts allow keys to be inserted
        for scope in [CommandScope.ENGINE, CommandScope.TRAIN, CommandScope.SWITCH, CommandScope.ACC]:
            cs_dict = ComponentStateDict(scope)
            for key in range(1, 99 + 1):
                assert key not in cs_dict
                value = cs_dict[key]
                assert key in cs_dict
                assert value.scope == scope
                assert isinstance(value, ComponentState)
                assert value.address == key
                assert value.last_command is None
                assert value.last_updated is None

        # verify invalid keys not allowed
        for scope in [CommandScope.TRAIN, CommandScope.SWITCH, CommandScope.ACC]:
            cs_dict = ComponentStateDict(scope)
            for key in [-4, 0, "abc", 100, CommandScope.TRAIN]:
                assert key not in cs_dict
                with pytest.raises(KeyError, match=f"Invalid ID: {key}"):
                    _ = cs_dict[key]
        for scope in [CommandScope.ENGINE]:
            cs_dict = ComponentStateDict(scope)
            for key in [-4, 0, "abc", 10000, CommandScope.TRAIN]:
                assert key not in cs_dict
                with pytest.raises(KeyError, match=f"Invalid ID: {key}"):
                    _ = cs_dict[key]
        for scope in [CommandScope.BASE]:
            cs_dict = ComponentStateDict(scope)
            for key in [-4, 1, "abc", 99, CommandScope.TRAIN]:
                assert key not in cs_dict
                with pytest.raises(KeyError, match=f"Invalid ID: {key}"):
                    _ = cs_dict[key]
        for scope in [CommandScope.SYNC]:
            cs_dict = ComponentStateDict(scope)
            for key in [-4, 1, "abc", 0, 32, CommandScope.TRAIN]:
                assert key not in cs_dict
                with pytest.raises(KeyError, match=f"Invalid ID: {key}"):
                    _ = cs_dict[key]

    def test_state_via_command_dispatcher(self) -> None:
        # set up a dispatcher and register ourselves as the handler
        dispatcher = CommandDispatcher()
        dispatcher.subscribe(self, CommandScope.ENGINE)
        dispatcher.subscribe(self, CommandScope.TRAIN)
        dispatcher.subscribe(self, CommandScope.SWITCH)
        dispatcher.subscribe(self, CommandScope.ACC)

        # verify state instance variable is empty
        assert not self.state

        # simulate a switch throw
        sw_out = CommandReq.build(Switch.OUT, 22)
        dispatcher.offer(sw_out)
        time.sleep(0.05)
        assert dispatcher.is_running
        assert self.state
        assert CommandScope.SWITCH in self.state
        assert 22 in self.state[CommandScope.SWITCH]
        assert self.state[CommandScope.SWITCH][22].last_command == sw_out
        assert self.state[CommandScope.SWITCH][22].address == 22
        assert self.state[CommandScope.SWITCH][22].scope == CommandScope.SWITCH
        assert self.state[CommandScope.SWITCH][22].last_updated is not None
        assert self.state[CommandScope.SWITCH][22].state == Switch.OUT
        assert self.state[CommandScope.SWITCH][22].is_known is True
        assert self.state[CommandScope.SWITCH][22].is_out is True
        assert self.state[CommandScope.SWITCH][22].is_through is False



from .tmcc2_command import TMCC2Command
from ..multibyte.multibyte_constants import TMCC2MultiByteEnum
from ..command_req import CommandReq
from ..constants import CommandScope, DEFAULT_BAUDRATE, DEFAULT_PORT
from .tmcc2_constants import TMCC2EngineCommandEnum


class EngineCmd(TMCC2Command):
    def __init__(
        self,
        engine: int,
        command: TMCC2EngineCommandEnum | TMCC2MultiByteEnum,
        data: int = 0,
        scope: CommandScope = CommandScope.ENGINE,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        if scope is None or scope not in [CommandScope.ENGINE, CommandScope.TRAIN]:
            raise ValueError(f"Scope must be ENGINE or TRAIN ({scope})")
        if scope == CommandScope.ENGINE:
            if engine < 1 or engine > 9999:
                raise ValueError(f"{scope.name.title()} must be between 1 and 9999")
        else:
            if engine < 1 or engine > 99:
                raise ValueError(f"{scope.name.title()} must be between 1 and 99")
        req = CommandReq.build(command, engine, data, scope)
        super().__init__(command, req, engine, data, scope, baudrate, port, server)



from __future__ import annotations

import sys
from enum import Enum, unique, IntEnum
from typing import Any, List, Dict, Set

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

"""
    General Constants
"""
DEFAULT_BAUDRATE: int = 9600
DEFAULT_VALID_BAUDRATES: List[int] = [9600, 19200, 38400, 57600, 115200]
DEFAULT_PORT: str = "/dev/ttyUSB0"
DEFAULT_ADDRESS: int = 99
BROADCAST_ADDRESS: int = 99
BROADCAST_TOPIC = "BROADCAST"
PROGRAM_NAME = "PyTrain"

DEFAULT_BASE_PORT: int = 50001

DEFAULT_SERVER_PORT: int = 5110  # unassigned by IANA as of 1/1/2025
DEFAULT_PULSE = 5  # send heartbeat periodically as proof of life

DEFAULT_QUEUE_SIZE: int = 2**12  # 4,096 entries

DEFAULT_SER2_THROTTLE_DELAY: int = 50  # milliseconds
DEFAULT_BASE_THROTTLE_DELAY: int = 50
DEFAULT_DURATION_INTERVAL_MSEC: int = 50
MINIMUM_DURATION_INTERVAL_MSEC: int = 20

DEFAULT_ENGINE_LABOR = 12

# AVAHI Service Constants
SERVICE_TYPE = f"_{PROGRAM_NAME.lower()}._tcp.local."
SERVICE_NAME = f"{PROGRAM_NAME}-Server.{SERVICE_TYPE}"


class Mixins(Enum):
    """
    Common mixins we want all PyLegacy enums to support
    """

    @classmethod
    def by_name(cls, name: str, raise_exception: bool = False) -> Self | None:
        if name is None:
            if raise_exception:
                raise ValueError(f"None is not a valid {cls.__name__}")
            else:
                return None
        orig_name = name = name.strip()
        if name in cls.__members__:
            return cls[name]
        # fall back to case-insensitive s
        name = name.upper()
        for k, v in cls.__members__.items():
            if k != k.upper():
                continue
            if k.upper() == name:
                return cls[k]
        if not raise_exception:
            return None
        if name:
            raise ValueError(f"'{orig_name}' is not a valid {cls.__name__}")
        else:
            raise ValueError(f"None/Empty is not a valid {cls.__name__}")

    @classmethod
    def by_prefix(cls, name: str, raise_exception: bool = False) -> Self | None:
        if name is None or not name.strip():
            if raise_exception:
                raise ValueError(f"None is not a valid {cls.__name__}")
            else:
                return None
        orig_name = name = name.strip()
        name = name.strip().upper()
        for k, v in cls.__members__.items():
            if k != k.upper():
                continue
            if k.upper().startswith(name):
                return cls[k]
        if not raise_exception:
            return None
        if name:
            raise ValueError(f"'{orig_name}' is not a valid {cls.__name__}")
        else:
            raise ValueError(f"None/Empty is not a valid {cls.__name__}")

    @classmethod
    def by_value(cls, value: Any, raise_exception: bool = False) -> Self | None:
        for _, member in cls.__members__.items():
            if member.value == value:
                return member
            if hasattr(member.value, "bits") and member.value.bits == value:
                return member
        if raise_exception:
            raise ValueError(f"'{value}' is not a valid {cls.__name__}")
        else:
            return None

    @classmethod
    def _missing_(cls, value) -> Self:
        if type(value) is str:
            value = str(value).upper()
            if value in dir(cls):
                return cls[value]
        elif type(value) is int:
            return cls.by_value(value, raise_exception=True)
        raise ValueError(f"{value} is not a valid {cls.__name__}")

    @property
    def title(self) -> str:
        return self.name.title()

    @property
    def label(self) -> str:
        return self.name.capitalize()


class OfficialRRSpeeds(Mixins, Enum):
    """
    Marker enum
    """


@unique
class CommandSyntax(Mixins, Enum):
    TMCC = 1
    LEGACY = 2
    PDI = 3


@unique
class CommandScope(Mixins, Enum):
    ENGINE = 1
    TRAIN = 2
    SWITCH = 3
    ROUTE = 4
    ACC = 5
    SYSTEM = 6
    ASC2 = 7
    AMC2 = 8
    BPC2 = 9
    IRDA = 10
    STM2 = 11
    BASE = 12
    SYNC = 13
    BLOCK = 14


@unique
class Direction(Mixins, Enum):
    UNKNOWN = 0
    L2R = 1
    R2L = 2


class CommandPrefix(Mixins, IntEnum):
    """
    Marker interface for Command Prefix enums
    """

    @property
    def prefix(self) -> Self:
        return self

    @property
    def as_int(self) -> int:
        return self.value

    @property
    def as_bytes(self) -> bytes:
        return self.as_int.to_bytes(1, byteorder="big")


"""
    Relative speed is specified with values ranging from -5 to 5 that are
    mapped to values 0 - 10
"""
RELATIVE_SPEED_MAP = dict(zip(range(-5, 6), range(0, 11)))
CONTROL_TYPE: Dict[int, str] = {
    0: "Cab-1",
    1: "TMCC",
    2: "Legacy",
    3: "R100",
}
SOUND_TYPE: Dict[int, str] = {
    0: "None",
    1: "RainSounds",
    2: "RailSounds 5",
    3: "Legacy",
}
LOCO_TYPE: Dict[int, str] = {
    0: "Diesel",
    1: "Steam",
    2: "Electric",
    3: "Subway",
    4: "Accessory/Operating Car",
    5: "Passenger",
    6: "Breakdown",
    7: "reserved",
    8: "Acela",
    9: "Track Crane",
    10: "Diesel Switcher",
    11: "Steam Switcher",
    12: "Freight",
    13: "Diesel Pullmor",
    14: "Steam Pullmor",
    15: "Transformer",
}
LOCO_CLASS: Dict[int, str] = {
    0: "Locomotive",
    1: "Switcher",
    2: "Subway",
    10: "Pullmor",
    20: "Transformer",
    255: "Universal",
}

LEGACY_CONTROL_TYPE = 2
CAB1_CONTROL_TYPE = 0
LOCO_ACCESSORY = 4
LOCO_TRACK_CRANE: int = 9
TRACK_CRANE_STATE_NUMERICS: Set[int] = {1, 2, 3}

RPM_TYPE = {0, 10, 13}
STEAM_TYPE = {1, 11, 14}


# Turn some of these into enums
class ControlTypeDef:
    def __init__(self, control_type: int, is_legacy: bool = False) -> None:
        self._control_type = control_type
        self._is_legacy = is_legacy

    @property
    def is_legacy(self) -> bool:
        return self._is_legacy

    @property
    def control_type(self) -> int:
        return self._control_type

    @property
    def bits(self) -> int:
        return self._control_type

    @property
    def label(self) -> str:
        return CONTROL_TYPE[self.control_type]


@unique
class ControlType(Mixins):
    CAB1 = ControlTypeDef(0)
    TMCC = ControlTypeDef(1)
    LEGACY = ControlTypeDef(2, True)
    R100 = ControlTypeDef(3)

    @property
    def is_legacy(self) -> bool:
        return self.value.is_legacy

    @property
    def label(self) -> str:
        return self.value.label



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import atexit
from threading import Event, RLock, Thread

from .. import ComponentStateStore, EngineState, TrainState
from ..db.state_watcher import StateWatcher
from ..protocol.constants import DEFAULT_ADDRESS, PROGRAM_NAME, CommandScope
from .gpio_device import GpioDevice
from .i2c.oled import Oled, OledDevice

UP = "\u25b4"
DOWN = "\u25be"
REV = "\u00ab"
FWD = "\u00bb"
BELL = "\u266b"


class EngineStatus(Thread, GpioDevice):
    def __init__(
        self,
        tmcc_id: int | EngineState = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        address: int = 0x3C,
        device: OledDevice | str = OledDevice.ssd1309,
    ) -> None:
        self._lock = RLock()
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Engine Status Oled")
        self._oled = Oled(address, device, auto_update=False)
        self._state_store = ComponentStateStore.get()
        if isinstance(tmcc_id, EngineState):
            self._monitored_state = tmcc_id
            self._tmcc_id = tmcc_id.address
            self._scope = tmcc_id.scope
        elif isinstance(tmcc_id, int) and 1 <= tmcc_id <= 9999 and scope in [CommandScope.ENGINE, CommandScope.TRAIN]:
            self._tmcc_id = tmcc_id
            self._scope = scope
            self._monitored_state = None
        else:
            raise ValueError(f"Invalid tmcc_id: {tmcc_id} or scope: {scope}")

        self._is_running = True
        self._ev = Event()
        self._railroad = None
        self._last_known_speed = self._monitored_state.speed if self._monitored_state else None
        self._state_watcher = None

        # check for state synchronization
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        if self._sync_state and self._sync_state.is_synchronized is True:
            self._sync_watcher = None
            self.on_sync()
        else:
            self.update_display()
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)
        atexit.register(self.close)

    @property
    def display(self) -> Oled:
        return self._oled

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @tmcc_id.setter
    def tmcc_id(self, value: int) -> None:
        self.update_engine(value, self.scope)

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @scope.setter
    def scope(self, value: CommandScope) -> None:
        self.update_engine(self.tmcc_id, value)

    @property
    def is_synchronized(self) -> bool:
        return self._synchronized

    @property
    def state(self) -> EngineState | TrainState | None:
        return self._monitored_state

    @property
    def railroad(self) -> str:
        if self._railroad is None:
            base_state = self._state_store.get_state(CommandScope.BASE, 0, False)
            if base_state and base_state.base_name:
                self._railroad = base_state.base_name.title()
        return self._railroad if self._railroad is not None else "Loading Engine Roster..."

    def run(self) -> None:
        while self._is_running is True and self._ev.is_set() is False:
            self._ev.wait(10)

    def update_engine(self, tmcc_id: int, scope: CommandScope = CommandScope.ENGINE) -> None:
        self._tmcc_id = tmcc_id
        self._scope = scope
        if tmcc_id is not None and tmcc_id != 99:
            self._monitored_state = self._state_store.get_state(scope, tmcc_id, create=False)
        else:
            self._monitored_state = None
        self._monitor_state_updates()
        self.update_display(clear=True)

    def update_display(self, clear: bool = False) -> None:
        with self._lock:
            if clear is True:
                self.display.clear()
            cursor = None
            if self._monitored_state:
                is_started = self._monitored_state.is_started
                is_shutdown = self._monitored_state.is_shutdown
                rname = self._monitored_state.road_name if self._monitored_state.road_name else "No Information"
                rnum = f"#{self._monitored_state.road_number} " if self._monitored_state.road_number else ""
                if self.display.cols <= 20:
                    ct = (
                        f" {self._monitored_state.control_type_label}"
                        if self._monitored_state.control_type_label
                        else ""
                    )
                else:
                    ct = ""
                lt = f" {self._monitored_state.engine_type_label}" if self._monitored_state.engine_type_label else ""
                self.display[0] = f"{rnum}{rname}{ct}{lt}"

                tmp = f"{self._scope.label}: "
                row = f"{tmp:<8}"
                row += f"{self._tmcc_id:04}"
                if self.display.cols > 20:
                    if rnum:
                        row += f" {rnum.strip()}"
                if self._monitored_state.control_type_label:
                    if self.display.cols > 15:
                        row += f" {self._monitored_state.control_type_label}"
                    else:
                        row += f" {self._monitored_state.control_type_label[0]}"
                self.display[1] = row

                row = f"Speed: {self._monitored_state.speed:03d}"
                if self.display.cols > 20:
                    row += f"/{self._monitored_state.speed_max:03d}"
                dr = self._monitored_state.direction_label
                if self.display.cols > 15:
                    if dr == "FW":
                        dr = "Fwd"
                    elif dr == "RV":
                        dr = "Rev"
                    else:
                        dr = "---"
                else:
                    if dr == "FW":
                        dr = "F" + FWD
                    elif dr == "RV":
                        dr = "R" + REV
                row += f" {dr}"
                if self.display.cols > 20:
                    row += " Started " + UP if is_started is True else " Shutdown" + DOWN if is_shutdown is True else ""
                elif self.display.cols > 15:
                    row += " Started" if is_started is True else " Off " + DOWN if is_shutdown is True else ""
                else:
                    row += UP if is_started is True else DOWN if is_shutdown is True else " "
                self.display[2] = row

                if self.display.cols > 20:
                    rpm = f" RPM: {self._monitored_state.rpm:1d}:{self._monitored_state.labor:02d}"
                else:
                    rpm = ""
                if self.display.cols > 15:
                    tb = f"TB: {self._monitored_state.train_brake}"
                    mo = f"Mo: {self._monitored_state.momentum}"
                    sm = f"Sm: {self._monitored_state.smoke_label if self._monitored_state.smoke_label else '?'}"
                else:
                    tb = f"B: {self._monitored_state.train_brake}"
                    mo = f"M: {self._monitored_state.momentum}"
                    sm = f"S: {self._monitored_state.smoke_label if self._monitored_state.smoke_label else '?'}"
                row = f"{tb} {mo} {sm}{rpm}"
                self.display[3] = row
                cursor = (1, 8)
            elif self.is_synchronized is True:
                self.display.write(self.railroad, 0, center=True)
                if self._tmcc_id and self._tmcc_id != 99:
                    self.display.write(f"{self.scope.label}: {self._tmcc_id}", 1, center=True)
                    self.display.write("Not Found", 2, center=True)
                    cursor = (1, len(self.display[1]) - len(str(self._tmcc_id)))
                else:
                    # display prompt for Engine/Train
                    self.display.write(f"{self.scope.label}: ?", 1, center=True)
                    cursor = (1, len(self.display[1]) - 1)
            else:
                self.display.write("Synchronizing...", 0, center=True, blink=True)
            self.display.update_display()
            if cursor is not None:
                self.display.cursor_pos = cursor

    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True
            if self._monitored_state is None and self.tmcc_id and self.tmcc_id != 99:
                self._monitored_state = self._state_store.get_state(self.scope, self.tmcc_id)
            self._monitor_state_updates()
            self.update_display(clear=True)

    def on_state_update(self) -> None:
        cur_speed = self._monitored_state.speed if self._monitored_state else None
        if cur_speed is not None and self._last_known_speed != cur_speed:
            self._last_known_speed = cur_speed
        self.update_display()

    def reset(self) -> None:
        self.display.reset()
        self._is_running = False
        self._ev.set()
        if self._sync_watcher:
            self._sync_watcher.shutdown()
            self._sync_watcher = None
        if self._state_watcher:
            self._state_watcher.shutdown()
            self._state_watcher = None

    def close(self) -> None:
        self.reset()

    def _monitor_state_updates(self):
        if self._state_watcher:
            self._state_watcher.shutdown()
            self._state_watcher = None

        if self._monitored_state:
            self._state_watcher = StateWatcher(self._monitored_state, self.on_state_update)



import logging
import time
from threading import Thread

import serial

from .command_listener import CommandListener
from ..protocol.constants import DEFAULT_BAUDRATE, DEFAULT_PORT, PROGRAM_NAME

log = logging.getLogger(__name__)


class SerialReader(Thread):
    def __init__(
        self, baudrate: int = DEFAULT_BAUDRATE, port: str = DEFAULT_PORT, consumer: CommandListener = None
    ) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Serial Port Reader")
        self._consumer = consumer
        self._baudrate = baudrate
        self._port = port
        self._is_running = True
        self.start()

    def run(self) -> None:
        with serial.Serial(
            self._port,
            self._baudrate,
            bytesize=serial.EIGHTBITS,
            exclusive=True,
            timeout=1.0,
        ) as ser:
            while self._is_running:
                try:
                    in_waiting = ser.in_waiting
                    if in_waiting > 0:
                        ser2_bytes = ser.read(256)
                        if ser2_bytes:
                            if self._consumer:
                                self._consumer.offer(ser2_bytes)
                            else:
                                log.warning(f"No serial consumer for: {ser2_bytes.hex(':')}")
                    # give the CPU a break
                    time.sleep(0.05)
                except serial.SerialException as se:
                    if se.errno is None:
                        pass
                    else:
                        log.exception(se)
                except Exception as e:
                    log.exception(e)

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @property
    def port(self) -> str:
        return self._port

    @property
    def is_consumer(self) -> bool:
        return self._consumer is not None

    @property
    def is_running(self) -> bool:
        return self._is_running

    def shutdown(self) -> None:
        self._is_running = False



from __future__ import annotations

from .constants import PdiCommand, Stm2Action, PDI_SOP, PDI_EOP
from .lcs_req import LcsReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum


class Stm2Req(LcsReq):
    def __init__(
        self,
        data: bytes | int,
        pdi_command: PdiCommand = PdiCommand.STM2_GET,
        action: Stm2Action = Stm2Action.CONFIG,
        ident: int | None = None,
        error: bool = False,
        mode: int = None,
        debug: int = None,
        state: TMCC1SwitchCommandEnum = None,
    ) -> None:
        super().__init__(data, pdi_command, action, ident, error)
        if isinstance(data, bytes):
            self._action = Stm2Action(self._action_byte)
            data_len = len(self._data)
            if self._action == Stm2Action.CONFIG:
                self._mode = self._data[7] if data_len > 7 else None
                self._debug = self._data[4] if data_len > 4 else None
            else:
                self._mode = self._debug = None

            if self._action == Stm2Action.CONTROL1:
                sw_state = self._data[3] if data_len > 3 else None
                self._state = TMCC1SwitchCommandEnum.OUT if sw_state == 1 else TMCC1SwitchCommandEnum.THRU
                self.scope = CommandScope.SWITCH
            else:
                self._state = None
        else:
            self._action = action
            self._mode = mode
            self._debug = debug
            self._state = state

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def debug(self) -> int | None:
        return self._debug

    @property
    def state(self) -> TMCC1SwitchCommandEnum:
        return self._state

    @property
    def is_thru(self) -> bool | None:
        return self._state == TMCC1SwitchCommandEnum.THRU

    @property
    def is_out(self) -> bool | None:
        return self._state == TMCC1SwitchCommandEnum.OUT

    @property
    def payload(self) -> str | None:
        if self.is_error:
            return super().payload
        if self.pdi_command != PdiCommand.STM2_GET:
            if self.action == Stm2Action.CONFIG:
                return f"Mode: {self.mode} Debug: {self.debug} ({self.packet})"
            elif self._action == Stm2Action.CONTROL1:
                return f"{self.state} ({self.packet})"
        return super().payload

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.tmcc_id.to_bytes(1, byteorder="big")
        byte_str += self.action.as_bytes

        if self._action == Stm2Action.CONFIG:
            if self.pdi_command != PdiCommand.STM2_GET:
                debug = self.debug if self.debug is not None else 0
                mode = self.mode if self.mode is not None else 0
                byte_str += self.tmcc_id.to_bytes(1, byteorder="big")  # allows board to be renumbered
                byte_str += debug.to_bytes(1, byteorder="big")
                byte_str += (0x0000).to_bytes(2, byteorder="big")
                byte_str += mode.to_bytes(1, byteorder="big")
        elif self._action == Stm2Action.IDENTIFY:
            if self.pdi_command == PdiCommand.STM2_SET:
                byte_str += (self.ident if self.ident is not None else 0).to_bytes(1, byteorder="big")
        elif self._action == Stm2Action.CONTROL1:
            if self.pdi_command != PdiCommand.STM2_GET:
                sw_state = 1 if self.state == TMCC1SwitchCommandEnum.OUT else 0
                byte_str += sw_state.to_bytes(1, byteorder="big")
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str



import unittest

from src.pytrain.gpio.i2c.mcp23017 import *
from .smbusmock import MBusMock


class TestMCP23017(unittest.TestCase):
    # noinspection PyTypeChecker
    i2c = I2C(MBusMock())
    mockAddress = 0x20

    def test_mcp23017_init(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        self.assertEqual(mcp.address, 0x20)

    def test_get_offset_pin_pair(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        pair = mcp.get_offset_gpio_tuple([GPINTENA, GPINTENB], GPA2)
        self.assertEqual(pair[0], GPINTENA)
        self.assertEqual(pair[1], GPA2)

        pair = mcp.get_offset_gpio_tuple([INTCONA, INTCONB], GPB0)
        self.assertEqual(pair[0], INTCONB)
        self.assertEqual(pair[1], GPB0 % 8)

    def test_get_offset_pin_pair_type_error(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        with self.assertRaises(TypeError):
            mcp.get_offset_gpio_tuple([INTCONA, INTCONB], 18)

        with self.assertRaises(TypeError):
            mcp.get_offset_gpio_tuple([INTCONA, INTCONB], -1)

        with self.assertRaises(TypeError):
            mcp.get_offset_gpio_tuple([0xFF, INTCONB], 1)

        with self.assertRaises(TypeError):
            mcp.get_offset_gpio_tuple([INTCONB, 0xFF], 1)

    def test_set_all_output(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_output()
        self.assertEqual(mcp.read(IODIRA), 0b00000000)
        self.assertEqual(mcp.read(IODIRB), 0b00000000)

    def test_set_all_input(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_input()
        self.assertEqual(mcp.read(IODIRA), 0b11111111)
        self.assertEqual(mcp.read(IODIRB), 0b11111111)

    def test_pin_mode(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_output()
        mcp.set_pin_mode(GPA0, INPUT)
        self.assertEqual(mcp.read(IODIRA), 0b00000001)
        mcp.set_pin_mode(GPA7, INPUT)
        self.assertEqual(mcp.read(IODIRA), 0b10000001)
        mcp.set_pin_mode(GPB7, INPUT)
        self.assertEqual(mcp.read(IODIRA), 0b10000001)
        self.assertEqual(mcp.read(IODIRB), 0b10000000)
        mcp.set_pin_mode(GPA0, OUTPUT)
        mcp.set_pin_mode(GPA7, OUTPUT)
        self.assertEqual(mcp.read(IODIRA), 0b00000000)

    def test_digital_write(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_output()
        self.assertEqual(mcp.read(OLATA), 0b00000000)
        self.assertEqual(mcp.read(OLATB), 0b00000000)

        mcp.digital_write(GPA0, HIGH)
        self.assertEqual(mcp.read(OLATA), 0b00000001)
        mcp.digital_write(GPA0, LOW)
        self.assertEqual(mcp.read(OLATA), 0b00000000)
        mcp.digital_write(GPA7, HIGH)
        self.assertEqual(mcp.read(OLATA), 0b10000000)

        mcp.digital_write(GPB0, HIGH)
        self.assertEqual(mcp.read(OLATA), 0b10000000)
        self.assertEqual(mcp.read(OLATB), 0b00000001)
        mcp.digital_write(GPB0, LOW)
        self.assertEqual(mcp.read(OLATA), 0b10000000)
        self.assertEqual(mcp.read(OLATB), 0b00000000)

    def test_set_interrupt(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_input()
        mcp.enable_interrupt(GPA0)
        self.assertEqual(mcp.read(GPINTENA), 0b00000001)
        mcp.enable_interrupt(GPA6)
        self.assertEqual(mcp.read(GPINTENA), 0b01000001)
        mcp.disable_interrupt(GPA0)
        self.assertEqual(mcp.read(GPINTENA), 0b01000000)
        mcp.enable_interrupt(GPB0)
        self.assertEqual(mcp.read(GPINTENB), 0b00000001)

    def test_set_all_interrupt(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_input()
        mcp.enable_interrupts()
        self.assertEqual(mcp.read(GPINTENA), 0b11111111)
        self.assertEqual(mcp.read(GPINTENB), 0b11111111)
        mcp.disable_interrupts()
        self.assertEqual(mcp.read(GPINTENA), 0b00000000)
        self.assertEqual(mcp.read(GPINTENB), 0b00000000)

    def test_interrupt_mirror(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_input()
        mcp.set_interrupt_mirror(True)
        self.assertEqual(mcp.read(IOCONA), 0b01000000)
        self.assertEqual(mcp.read(IOCONB), 0b01000000)
        mcp.set_interrupt_mirror(False)
        self.assertEqual(mcp.read(IOCONA), 0b00000000)
        self.assertEqual(mcp.read(IOCONB), 0b00000000)

    def test_read_interrupt_captures(self):
        mcp = Mcp23017(self.mockAddress, self.i2c)
        mcp.set_all_output()
        self.assertEqual(mcp.read(INTCAPA), 0b00000000)
        self.assertEqual(mcp.read(INTCAPB), 0b00000000)
        mcp.write(INTCAPA, 0x01)
        interrupt_captures = mcp.read_interrupt_captures()
        self.assertEqual(interrupt_captures[0][0], "1")
